"""
Overview:
    Import functions, may be useful when support some dynamic features (such as imports in cli development).
"""
import builtins
import fnmatch
import importlib
from itertools import islice
from queue import Queue
from typing import Optional, Callable, Any, Tuple, Iterator

from .func import dynamic_call

__all__ = [
    'import_object',
    'quick_import_object',
    'iter_import_objects',
]


def _import_module(module_name=None):
    if module_name:
        return importlib.import_module(module_name)
    else:
        return builtins


def import_object(obj_name: str, module_name: Optional[str] = None):
    """
    Overview:
        Dynamically import an object from module.

    Arguments:
        - obj_name (:obj:`str`): Name of the object.
        - module_name (:obj:`Optional[str]`): Name of the module, \
            default is ``None`` which means the ``builtins`` module.

    Returns:
        - obj: Imported object.

    Example::
        >>> import_object('zip')               # <class 'zip'>
        >>> import_object('ndarray', 'numpy')  # <class 'numpy.ndarray'>
    """
    return getattr(_import_module(module_name), obj_name)


def quick_import_object(full_name: str, predicate: Optional[Callable] = None) -> Tuple[Any, str, str]:
    """
    Overview:
        Quickly dynamically import an object with a single name.

    Arguments:
        - full_name (:obj:`str`): Full name of the object, attribute is supported as well.
        - predicate (:obj:`Callable`): Predicate function, default is ``None`` means no limitation.

    Returns:
        - obj (:obj:`Tuple[Any, str, str]`): Imported object.

    Example::
        >>> quick_import_object('zip')                     # <class 'zip'>, '', 'zip'
        >>> quick_import_object('numpy.ndarray')           # <class 'numpy.ndarray'>, 'numpy', 'ndarray'
        >>> quick_import_object('numpy.ndarray.__name__')  # 'ndarray', 'numpy', 'ndarray.__name__'
    """
    _iter = islice(iter_import_objects(full_name, predicate), 1)

    try:
        # noinspection PyTupleAssignmentBalance
        _obj, _module, _name = next(_iter)
        return _obj, _module, _name
    except (StopIteration, StopAsyncIteration):
        raise ImportError(f'Cannot import object {repr(full_name)}.')


def iter_import_objects(full_pattern: str, predicate: Optional[Callable] = None) \
        -> Iterator[Tuple[Any, str, str]]:
    """
    Overview:
        Quickly dynamically import all objects with full name pattern.

    Arguments:
        - full_pattern (:obj:`str`): Full pattern of the object, attribute is supported as well.
        - predicate (:obj:`Callable`): Predicate function, default is ``None`` means no limitation.

    Returns:
        - iter (:obj:`Iterator[Tuple[Any, str, str]]`): Iterator for all the imported objects.
    """
    predicate = dynamic_call(predicate or (lambda: True))

    segments = full_pattern.split('.')
    length = len(segments)
    _errs = []
    for i in reversed(range(length + 1)):
        module_name = '.'.join(segments[:i])
        attrs = tuple(segments[i:])
        attrs_count = len(attrs)

        try:
            module = importlib.import_module(module_name or 'builtins')
        except (ModuleNotFoundError, ImportError):
            continue

        queue = Queue()
        queue.put((module, 0, ()))
        exist = False

        while not queue.empty():
            root, pos, ats = queue.get()

            if pos >= attrs_count:
                obj_name = '.'.join(ats)
                if predicate(root, module_name, obj_name):
                    yield root, module_name, obj_name
            elif hasattr(root, attrs[pos]):
                queue.put((getattr(root, attrs[pos]), pos + 1, ats + (attrs[pos],)))
                exist = True
            elif hasattr(root, '__dict__'):
                for key, value in sorted(root.__dict__.items()):
                    if fnmatch.fnmatch(key, attrs[pos]):
                        queue.put((value, pos + 1, ats + (key,)))
                        exist = True

        if exist:
            break



from typing import Iterator, Mapping, Optional, List, Any

from .base import BaseGenerator, _single_dict_process, _single_to_tuple, _check_keys


class MatrixGenerator(BaseGenerator):
    """
    Full matrix model, all the cases in this matrix will be iterated.
    """

    def __init__(self, values: Mapping[str, Any],
                 names: Optional[List[str]] = None,
                 includes: Optional[List[Mapping[str, Any]]] = None,
                 excludes: Optional[List[Mapping[str, Any]]] = None):
        """
        Constructor of the :class:`hbutils.testing.MatrixGenerator` class.
        It is similar to GitHub Action's matrix.

        :param values: Matrix values, such as ``{'a': [2, 3], 'b': ['b', 'c']}``.
        :param names: Names of the given generator, default is ``None`` which means use the sorted \
            key set of the values.
        :param includes: Include items, such as ``[{'a': 4, 'b': 'b'}]``, \
            default is ``None`` which means no extra inclusions.
        :param excludes: Exclude Items, such as ``[{'a': 2, 'b': 'c'}]``, \
            default is ``None`` which means no extra exclusions.
        """
        BaseGenerator.__init__(self, values, names)
        _name_set = set(self.names)

        self.__includes = [_single_dict_process(inc) for inc in (includes or [])]
        for includes in self.__includes:
            _check_keys(includes, _name_set)

        self.__excludes = [_single_dict_process(exc) for exc in (excludes or [])]
        for excludes in self.__excludes:
            _check_keys(excludes, _name_set)

    @property
    def includes(self) -> List[Mapping[str, Any]]:
        """
        Include items.
        """
        return self.__includes

    @property
    def excludes(self) -> List[Mapping[str, Any]]:
        """
        Exclude items.
        """
        return self.__excludes

    def cases(self) -> Iterator[Mapping[str, Any]]:
        """
        Get the cases in this matrix.

        Examples::
            >>> from hbutils.testing import MatrixGenerator
            >>> for p in MatrixGenerator(
            ...         {'a': [1, 2, 3], 'b': ['a', 'b'], 'r': [3, 4, 5]},
            ...         includes=[{'a': 4, 'r': 7}],
            ...         excludes=[{'a': 1, 'r': 3}, {'a': 3, 'b': 'b'}, {'a': 4, 'b': 'a', 'r': 7}]
            ... ).cases():
            >>>     print(p)
            {'a': 1, 'b': 'a', 'r': 4}
            {'a': 1, 'b': 'a', 'r': 5}
            {'a': 1, 'b': 'b', 'r': 4}
            {'a': 1, 'b': 'b', 'r': 5}
            {'a': 2, 'b': 'a', 'r': 3}
            {'a': 2, 'b': 'a', 'r': 4}
            {'a': 2, 'b': 'a', 'r': 5}
            {'a': 2, 'b': 'b', 'r': 3}
            {'a': 2, 'b': 'b', 'r': 4}
            {'a': 2, 'b': 'b', 'r': 5}
            {'a': 3, 'b': 'a', 'r': 3}
            {'a': 3, 'b': 'a', 'r': 4}
            {'a': 3, 'b': 'a', 'r': 5}
            {'a': 4, 'b': 'b', 'r': 7}

        """
        n = len(self.names)

        def _check_single_exclude(dict_value, exclude):
            for key, value in exclude.items():
                if key not in dict_value or dict_value[key] not in value:
                    return False

            return True

        def _check_exclude(dict_value, excludes):
            for exclude in excludes:
                if _check_single_exclude(dict_value, exclude):
                    return True

            return False

        def _matrix_recursion(depth, dict_value, values, excludes):
            if _check_exclude(dict_value, excludes):
                return

            if depth < n:
                name = self.names[depth]
                for curitem in values[name]:
                    yield from _matrix_recursion(depth + 1, {**dict_value, name: curitem}, values, excludes)
            else:
                yield dict_value

        value_items = [self.values, *({
            name: _single_to_tuple(include[name]) if name in include else self.values[name]
            for name in self.names
        } for include in self.includes)]
        local_excludes = [*self.excludes]
        for vis in value_items:
            yield from _matrix_recursion(0, {}, vis, local_excludes)
            local_excludes.append(vis)



# hbutils.model's package info
_PACKAGE_NAME = 'hbutils.model'
_PACKAGE_HOMEPAGE = 'https://hansbug.github.io/hbutils/main/api_doc/model/index.html'
_PACKAGE_RST = f'`{_PACKAGE_NAME} <{_PACKAGE_HOMEPAGE}>`_'



"""
Overview:
    Utilities for processing streams.
"""
import io
import os
from contextlib import contextmanager
from typing import Union, TextIO, BinaryIO, ContextManager

__all__ = [
    'keep_cursor', 'getsize', 'is_eof',
]


@contextmanager
def keep_cursor(file: Union[TextIO, BinaryIO]) -> ContextManager:
    """
    Overview:
        Keep the cursor of the given file within a with-block.

    :param file: File which cursor need to be kept.

    Examples::
        >>> import io
        >>> from hbutils.file import keep_cursor
        >>>
        >>> with io.BytesIO(b'\\xde\\xad\\xbe\\xef') as file:
        ...     with keep_cursor(file):
        ...         print(file.read(2))
        ...     with keep_cursor(file):  # still from 0
        ...         print(file.read())
        ...
        ...     _ = file.read(2)
        ...     with keep_cursor(file):  # now from 2
        ...         print(file.read(1))
        ...     with keep_cursor(file):  # still from 2
        ...         print(file.read())
        b'\\xde\\xad'
        b'\\xde\\xad\\xbe\\xef'
        b'\\xbe'
        b'\\xbe\\xef'

    .. note::
        Only seekable stream can use :func:`keep_cursor`.
    """
    if file.seekable():
        curpos = file.tell()
        try:
            yield
        finally:
            file.seek(curpos, io.SEEK_SET)
    else:
        raise OSError(f'Given file {repr(file)} is not seekable, '  # pragma: no cover
                      f'so its cursor position cannot be kept.')


def getsize(file: Union[TextIO, BinaryIO]) -> int:
    """
    Overview:
        Get the size of the given ``file`` stream.

    :param file: File which size need to access.
    :return: File's size.

    Examples::
        >>> import io
        >>> from hbutils.file import getsize
        >>>
        >>> with io.BytesIO(b'\\xde\\xad\\xbe\\xef') as file:
        ...     print(getsize(file))
        4
        >>> with open('README.md', 'r') as file:
        ...     print(getsize(file))
        2582

    .. note::
        Only seekable stream can use :func:`getsize`.
    """
    if file.seekable():
        try:
            return os.stat(file.fileno()).st_size
        except OSError:
            with keep_cursor(file):
                return file.seek(0, io.SEEK_END)
    else:
        raise OSError(f'Given file {repr(file)} is not seekable, '  # pragma: no cover
                      f'so its size is unavailable.')


def is_eof(file: Union[TextIO, BinaryIO]) -> bool:
    """
    Overview:
        Check if the file meets its end.

    :param file: File to be checked.
    :return: Is EOF(end of file) or not.

    Examples::
        >>> import io
        >>> from hbutils.file import is_eof
        >>>
        >>> with io.BytesIO(b'\\xde\\xad\\xbe\\xef') as file:
        ...     print(file.tell(), is_eof(file))
        ...     _ = file.read(2)
        ...     print(file.tell(), is_eof(file))
        ...     _ = file.read(2)
        ...     print(file.tell(), is_eof(file))
        0 False
        2 False
        4 True
        >>> with open('README.md', 'r') as file:
        ...     print(file.tell(), is_eof(file))
        ...     _ = file.read(100)
        ...     print(file.tell(), is_eof(file))
        ...     _ = file.read()
        ...     print(file.tell(), is_eof(file))
        0 False
        100 False
        2582 True

    .. note::
        Only seekable stream can use :func:`is_eof`.
    """
    return file.tell() == getsize(file)



import functools
import itertools
import os
import pathlib
import subprocess
import sys
from typing import List, Optional

from packaging.requirements import Requirement
from packaging.utils import canonicalize_name

try:
    import importlib.metadata as importlib_metadata
except (ModuleNotFoundError, ImportError):
    import importlib_metadata
from packaging.version import Version

__all__ = [
    'package_version',
    'load_req_file', 'pip',
    'check_reqs', 'check_req_file',
    'pip_install', 'pip_install_req_file',
]


def package_version(name: str) -> Optional[Version]:
    """
    Overview:
        Get version of package with given ``name``.

    :param name: Name of the package, case is not sensitive.
    :return: A :class:`packing.version.Version` object. If the package is not installed, return ``None``.

    Examples::
        >>> from hbutils.system import package_version
        >>>
        >>> package_version('pip')
        <Version('21.3.1')>
        >>> package_version('setuptools')
        <Version('59.6.0')>
        >>> package_version('not_a_package')
        None
    """
    try:
        return Version(importlib_metadata.distribution(canonicalize_name(name)).version)
    except importlib_metadata.PackageNotFoundError:
        return None


def _nonblank(text):
    return text and not text.startswith('#')


@functools.singledispatch
def yield_lines(iterable):
    r"""
    Based on https://github.com/jaraco/jaraco.text/blob/main/jaraco/text/__init__.py#L537 .
    Yield valid lines of a string or iterable.
    >>> list(yield_lines(''))
    []
    >>> list(yield_lines(['foo', 'bar']))
    ['foo', 'bar']
    >>> list(yield_lines('foo\nbar'))
    ['foo', 'bar']
    >>> list(yield_lines('\nfoo\n#bar\nbaz #comment'))
    ['foo', 'baz #comment']
    >>> list(yield_lines(['foo\nbar', 'baz', 'bing\n\n\n']))
    ['foo', 'bar', 'baz', 'bing']
    """
    return itertools.chain.from_iterable(map(yield_lines, iterable))


@yield_lines.register(str)
def _(text):
    return filter(_nonblank, map(str.strip, text.splitlines()))


def drop_comment(line):
    """
    Based on https://github.com/jaraco/jaraco.text/blob/main/jaraco/text/__init__.py#L560 .
    Drop comments.
    >>> drop_comment('foo # bar')
    'foo'
    A hash without a space may be in a URL.
    >>> drop_comment('https://example.com/foo#bar')
    'https://example.com/foo#bar'
    """
    return line.partition(' #')[0]


def join_continuation(lines):
    r"""
    Based on https://github.com/jaraco/jaraco.text/blob/main/jaraco/text/__init__.py#L575 .
    Join lines continued by a trailing backslash.
    >>> list(join_continuation(['foo \\', 'bar', 'baz']))
    ['foobar', 'baz']
    >>> list(join_continuation(['foo \\', 'bar', 'baz']))
    ['foobar', 'baz']
    >>> list(join_continuation(['foo \\', 'bar \\', 'baz']))
    ['foobarbaz']
    Not sure why, but...
    The character preceding the backslash is also elided.
    >>> list(join_continuation(['goo\\', 'dly']))
    ['godly']
    A terrible idea, but...
    If no line is available to continue, suppress the lines.
    >>> list(join_continuation(['foo', 'bar\\', 'baz\\']))
    ['foo']
    """
    lines = iter(lines)
    for item in lines:
        while item.endswith('\\'):
            try:  # pragma: no cover
                item = item[:-2].strip() + next(lines)
            except StopIteration:
                return
        yield item


def load_req_file(requirements_file: str) -> List[str]:
    """
    Overview:
        Load requirements items from a ``requirements.txt`` file.

    :param requirements_file: Requirements file.
    :return requirements: List of requirements.

    Examples::
        >>> from hbutils.system import load_req_file
        >>> load_req_file('requirements.txt')
        ['packaging>=21.3', 'setuptools>=50.0']
    """
    with pathlib.Path(requirements_file).open() as reqfile:
        return list(map(
            lambda x: str(Requirement(x)),
            join_continuation(map(drop_comment, yield_lines(reqfile)))
        ))


def pip(*args, silent: bool = False):
    """
    Overview:
        Run pip command with code.

    :param args: Command line arguments for ``pip`` command.
    :param silent: Do not print anything. Default is false, which means print the output to ``sys.stdout`` \
        and ``sys.stderr``.

    Examples::
        >>> from hbutils.system import pip
        >>> pip('-V')
        pip 22.3.1 from /home/user/myproject/venv/lib/python3.7/site-packages/pip (python 3.7)
        >>> pip('-V', silent=True)  # nothing will be printed
    """
    process = subprocess.run(
        [sys.executable, '-m', 'pip', *args],
        stdin=sys.stdin if not silent else None,
        stdout=sys.stdout if not silent else subprocess.PIPE,
        stderr=sys.stderr if not silent else subprocess.PIPE,
    )
    assert not process.returncode, f'Error when calling {process.args!r}{os.linesep}' \
                                   f'Error Code - {process.returncode}{os.linesep}' \
                                   f'Stdout:{os.linesep}' \
                                   f'{process.stdout.decode()}{os.linesep}' \
                                   f'{os.linesep}' \
                                   f'Stderr:{os.linesep}' \
                                   f'{process.stderr.decode()}{os.linesep}'
    process.check_returncode()


def _yield_reqs_to_install(req: Requirement, current_extra: str = ''):
    if req.marker and not req.marker.evaluate({'extra': current_extra}):
        return

    try:
        version = importlib_metadata.distribution(req.name).version
    except importlib_metadata.PackageNotFoundError:  # req not installed
        yield req
    else:
        if req.specifier.contains(version):
            for child_req in (importlib_metadata.metadata(req.name).get_all('Requires-Dist') or []):
                child_req_obj = Requirement(child_req)

                need_check, ext = False, None
                for extra in req.extras:
                    if child_req_obj.marker and child_req_obj.marker.evaluate({'extra': extra}):
                        need_check = True
                        ext = extra
                        break

                if need_check:  # check for extra reqs
                    yield from _yield_reqs_to_install(child_req_obj, ext)

        else:  # main version not match
            yield req


def _check_req(req: Requirement):
    return not bool(list(itertools.islice(_yield_reqs_to_install(req), 1)))


def check_reqs(reqs: List[str]) -> bool:
    """
    Overview:
        Check if the given requirements are all satisfied.

    :param reqs: List of requirements.
    :return satisfied: All the requirements in ``reqs`` satisfied or not.

    Examples::
        >>> from hbutils.system import check_reqs
        >>> check_reqs(['pip>=20.0'])
        True
        >>> check_reqs(['pip~=19.2'])
        False
        >>> check_reqs(['pip>=20.0', 'setuptools>=50.0'])
        True

    .. note::
        If a requirement's marker is not satisfied in this environment,
        **it will be ignored** instead of return ``False``.
    """
    return all(map(lambda x: _check_req(Requirement(x)), reqs))


def check_req_file(requirements_file: str) -> bool:
    """
    Overview:
        Check if the requirements in the given ``requirements_file`` is satisfied.

    :param requirements_file: Requirements file, such as ``requirements.txt``.
    :return satisfied: All the requirements in ``requirements_file`` satisfied or not.

    Examples::
        >>> from hbutils.system import check_req_file
        >>>
        >>> check_req_file('requirements.txt')
        True
        >>> check_req_file('requirements-test.txt')
        True
    """
    return check_reqs(load_req_file(requirements_file))


def pip_install(reqs: List[str], silent: bool = False, force: bool = False, user: bool = False):
    """
    Overview:
        Pip install requirements with code.
        Similar to ``pip install req1 req2 ...``.

    :param reqs: Requirement items to install.
    :param silent: Do not print anything. Default is ``False``.
    :param force: Force execute the ``pip install`` command. Default is ``False`` which means the requirements \
        will be checked before installation, and the installation will be only executed when \
        some requirements not installed.
    :param user: User mode, represents ``--user`` option in ``pip``.

    Examples::
        >>> from hbutils.system import pip_install
        >>> pip_install(['scikit-learn'])  # not installed
        Looking in indexes: https://xxx/simple
        Collecting scikit-learn
          Using cached https://xxx/scikit_learn-1.0.2-cp37-cp37m-manylinux_2_17_x86_64.manylinux2014_x86_64.whl (24.8 MB)
        Installing collected packages: threadpoolctl, scipy, joblib, scikit-learn
        Successfully installed joblib-1.2.0 scikit-learn-1.0.2 scipy-1.7.3 threadpoolctl-3.1.0
        >>> pip_install(['numpy>=1.10.0'])  # installed
        >>> pip_install(['numpy>=1.10.0'], force=True)  # force execute
        Looking in indexes: https://xxx/simple
        Requirement already satisfied: numpy>=1.10.0 in ./venv/lib/python3.7/site-packages (1.21.6)
    """
    if force or not check_reqs(reqs):
        pip('install', *(('--user',) if user else ()), *reqs, silent=silent)


def pip_install_req_file(requirements_file: str, silent: bool = False, force: bool = False, user: bool = False):
    """
    Overview:
        Pip install requirements from file with code.
        Similar to ``pip install -r requirements.txt``.

    :param requirements_file: Requirements file, such as ``requirements.txt``.
    :param silent: Do not print anything. Default is ``False``.
    :param force: Force execute the ``pip install`` command. Default is ``False`` which means the requirements \
        will be checked before installation, and the installation will be only executed when \
        some requirements not installed.
    :param user: User mode, represents ``--user`` option in ``pip``.

    Examples::
        >>> from hbutils.system import pip_install_req_file
        >>> pip_install_req_file('requirements.txt')  # pip install -r requirements.txt
    """
    if force or not check_req_file(requirements_file):
        pip('install', *(('--user',) if user else ()), '-r', requirements_file, silent=silent)



"""
Overview:
    Capture for exitcode.
"""
from contextlib import contextmanager
from typing import ContextManager
from unittest.mock import patch

__all__ = [
    'capture_exit', 'ExitCaptureResult'
]

_DEFAULT_EXITCODE = 0


def _exitcode(status=None):
    return status if status is not None else _DEFAULT_EXITCODE


def _fake_exit(status=None):
    raise SystemExit(status)


class ExitCaptureResult:
    """
    Overview:
        Model of exit capture result.
    """

    def __init__(self, exitcode):
        """
        Constructor of :class:`ExitCaptureResult`.

        :param exitcode: Exitcode value.
        """
        self.__exitcode = exitcode

    @property
    def exitcode(self) -> int:
        """
        Exitcode value.

        .. note::
            Do not use this property when :func:`capture_exit` is not over, otherwise this result \
                is not guaranteed to be correct.
        """
        return self.__exitcode

    def put_result(self, exitcode: int):
        """
        Put result inside this model.

        :param exitcode: New exitcode value.
        """
        self.__exitcode = exitcode


@contextmanager
def capture_exit(default=0) -> ContextManager[ExitCaptureResult]:
    """
    Overview:
        Capture for exitcode, :func:`quit` and :func:`sys.exit` can be captured.

    :param default: Default exitcode, default is ``0``.

    Examples::
        >>> from hbutils.testing import capture_exit
        >>> with capture_exit() as ex:
        ...     pass
        >>> ex.exitcode
        0
        >>>
        >>> with capture_exit() as ex:
        ...     quit()
        >>> ex.exitcode
        0
        >>>
        >>> with capture_exit() as ex:
        ...     quit(0xf)
        >>> ex.exitcode
        15
        >>>
        >>> import sys
        >>> with capture_exit() as ex:
        ...     sys.exit(0xf)
        >>> ex.exitcode
        15
    """
    capture = ExitCaptureResult(default)
    try:
        with patch('sys.exit', _fake_exit):
            yield capture
    except SystemExit as err:
        capture.put_result(_exitcode(err.code))



"""
Overview:
    Useful utilities for truncate (shorten) your string.
"""
from textwrap import shorten as _shorten

__all__ = ['truncate']


def truncate(text: str, width: int = 70, tail_length: int = 0, show_length: bool = False):
    """
    Overview:
        Truncate string into short form.

    Arguments:
        - text (:obj:`str`): Original text to be truncated.
        - width (:obj:`int`): Final width of the new string, default is ``70``.
        - tail_length (:obj:`int`): Tail's length of the new string, \
            default is ``0`` which means no tail.
        - show_length (:obj:`bool`): Show length in middle part or not, \
            default is ``False`` which means do not show this.

    Returns:
        - shortened (:obj:`str`) Short-formed string.

    Examples::
        >>> from hbutils.string import truncate
        >>> truncate('abc ' * 30, width=30)
        'abc abc abc abc abc abc ... '
        >>> truncate('abc ' * 30, width=40, tail_length=10)
        'abc abc abc abc abc abc ... c abc abc '
        >>> truncate('abc ' * 30, width=40, tail_length=10, show_length=True)
        'abc abc abc ..(120 chars).. c abc abc '

    """
    text = str(text)

    if show_length:
        omission = ' ..({length} chars).. '.format(length=len(text))
    else:
        omission = ' ... '
    if tail_length:
        omission += text[len(text) - tail_length:]

    return _shorten(text, width=width, placeholder=omission)



import socket
import time
from telnetlib import Telnet

__all__ = [
    'telnet', 'wait_for_port_online',
]

from typing import Optional


def telnet(host, port: int, timeout: float = 5.0) -> bool:
    """
    Overview:
        Perform a telnet operation on the given port and return ``True`` if a response is received
        within the timeout period, otherwise return ``False``.

    :param host: Host, e.g. ``127.0.0.1``.
    :param port: Port to test, e.g. ``32768``.
    :param timeout: Maximum timeout duration in seconds.
    :return: Port is on or not.

    """
    try:
        with Telnet(host, port, timeout):
            return True
    except (ConnectionRefusedError, socket.timeout):
        # ConnectionRefusedError is for Linux and macOS
        # socket.timeout is for Windows
        return False


def wait_for_port_online(host, port: int, timeout: Optional[float] = 5, interval: float = 0.3):
    """
    Overview:
        Wait for the given interface to be in an online state.

    :param host: Host, e.g. ``127.0.0.1``.
    :param port: Port to test, e.g. ``32768``.
    :param timeout: Maximum timeout duration in seconds. ``None`` means wait it forever.
    :param interval: Timeout for every :func:`telnet` operation.
    :raises TimeoutError: Raise this when timeout is reached and the port is still offline.
    """
    _start_time = time.time()
    while True:
        if telnet(host, port, interval):
            return

        if timeout is not None and time.time() > _start_time + timeout:
            raise TimeoutError(f'Wait for {host}:{port} for {time.time() - _start_time:.3f}s, '
                               f'but still offline.')



"""
Overview:
    Useful functions for processing python classes and types.
"""
from functools import WRAPPER_ASSIGNMENTS as CLASS_WRAPPER_ASSIGNMENTS
from functools import update_wrapper, partial
from typing import Tuple

__all__ = [
    'class_wraps',
    'common_base',
]

CLASS_WRAPPER_UPDATES = ()


def class_wraps(wrapped: type,
                assigned: Tuple[str] = CLASS_WRAPPER_ASSIGNMENTS,
                updated: Tuple[str] = CLASS_WRAPPER_UPDATES):
    r"""
    Overview:
        Wrapper decorator for class.

    Arguments:
        - wrapped (:obj:`type`): Wrapped class.
        - assigned (:obj:`Tuple[str]`): Wrapper assignments, equal to :func:`functools.wraps`'s \
            ``WRAPPER_ASSIGNMENTS``.
        - updated (:obj:`Tuple[str]`): Wrapper updates, default is ``()``, no update will be done.

    Examples:
        >>> def cls_dec(clazz):
        >>>     @class_wraps(clazz)
        >>>     class _NewClazz(clazz):
        >>>         pass
        >>>
        >>>     return _NewClazz
    """
    return partial(update_wrapper, wrapped=wrapped,
                   assigned=assigned, updated=updated)


def common_base(cls: type, *clss: type) -> type:
    """
    Overview:
        Get common base class of the given classes.
        Only ``__base__`` is considered.

    Arguments:
        - cls (:obj:`type`): First class.
        - clss (:obj:`type`): Other classes.

    Returns:
        - base (:obj:`type`): Common base class.

    Examples::
        >>> from hbutils.reflection import common_base
        >>> common_base(object)
        <class 'object'>
        >>> common_base(object, int, str)
        <class 'object'>
        >>> common_base(RuntimeError, ValueError, KeyError)
        <class 'Exception'>
    """
    current_cls = cls
    for new_cls in clss:
        while not issubclass(new_cls, current_cls):
            current_cls = current_cls.__base__

        if current_cls is object:
            break

    return current_cls



"""
Overview:
    Functions to check if a file is binary or text.
"""
__all__ = [
    'is_binary_file', 'is_text_file',
]

_TEXT_CHARS = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7f})


def _is_binary_string(data: bytes) -> bool:
    return bool(data.translate(None, _TEXT_CHARS))


def _take_sample(filename, size=1024) -> bytes:
    with open(filename, 'rb') as f:
        return f.read(size)


def is_binary_file(filename) -> bool:
    """
    Overview:
        Check if the given file is binary file.

    :param filename: Filename.
    :return: Is binary file or not.

    Examples::
        >>> from hbutils.system import is_binary_file
        >>> is_binary_file('rar_template-simple.rar')
        True
        >>> is_binary_file('README.md')
        False

    .. note::
        Empty file will be seen as text file.
    """
    return _is_binary_string(_take_sample(filename))


def is_text_file(filename) -> bool:
    """
    Overview:
        Check if the given file is text file.

    :param filename: Filename.
    :return: Is text file or not.
    
    Examples::
        >>> from hbutils.system import is_text_file
        >>> is_text_file('rar_template-simple.rar')
        False
        >>> is_text_file('README.md')
        True

    .. note::
        Empty file will be seen as text file.
    """
    return not _is_binary_string(_take_sample(filename))



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Jonathan M. Lange <jml@mumak.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This is a copy of https://github.com/jml/tree-format, with a few modifications,
# its based commit id is 4c6de1074d96129b7e03eecdf42fac2cde3b5151.

r"""
Overview:
    Library for formatting trees.

    This is a copy of https://github.com/jml/tree-format, with a few modifications,
    its based commit id is 4c6de1074d96129b7e03eecdf42fac2cde3b5151.

Changes:
    - The ``NEWLINE`` value is modified to empty string for the vision of final tree.
    - The ``print_tree`` function is removed because it is nowhere to be used in our case.
    - Add ``__doc__`` for ``format_tree`` function.
    - All the ``\n`` strings are replaced to ``os.linesep``.
    - The code is reformatted.
"""

import itertools
import os
import sys

__all__ = [
    'format_tree',
]

_DEFAULT_ENCODING = os.environ.get("PYTHONIOENCODING", sys.getdefaultencoding())

_UTF8_CHARS = (u'\u251c', u'\u2514', u'\u2502', u'\u2500', u'')
_ASCII_CHARS = (u'+', u'`', u'|', u'-', u'')


def _format_newlines(prefix, formatted_node, chars: tuple):
    """
    Convert newlines into U+23EC characters, followed by an actual newline and
    then a tree prefix so as to position the remaining text under the previous
    line.
    """
    FORK, LAST, VERTICAL, HORIZONTAL, NEWLINE = chars
    replacement = u''.join([NEWLINE, os.linesep, prefix])
    return replacement.join(formatted_node.splitlines())


def _format_tree(node, format_node, get_children, prefix=u'', chars: tuple = _UTF8_CHARS):
    FORK, LAST, VERTICAL, HORIZONTAL, NEWLINE = chars
    children = list(get_children(node))
    next_prefix = u''.join([prefix, VERTICAL, u'   '])
    for child in children[:-1]:
        yield u''.join([
            prefix, FORK, HORIZONTAL, HORIZONTAL, u' ',
            _format_newlines(next_prefix, format_node(child), chars)])
        for result in _format_tree(child, format_node, get_children, next_prefix, chars):
            yield result
    if children:
        last_prefix = u''.join([prefix, u'    '])
        yield u''.join([
            prefix, LAST, HORIZONTAL, HORIZONTAL, u' ',
            _format_newlines(last_prefix, format_node(children[-1]), chars)])
        for result in _format_tree(children[-1], format_node, get_children, last_prefix, chars):
            yield result


def format_tree(node, format_node, get_children, encoding=None) -> str:
    r"""
    Overview:
        Format the given tree.

    Arguments:
        - node: Node object
        - format_node: Format node getter
        - get_children: Children getter.
        - encoding: Encoding to be used. Default is ``None`` which means system encoding. \
            When ASCII encoding is used, ASCII chars will be used instead of original chars.

    Returns:
        - formatted: Formatted string.

    Example:
        >>> from operator import itemgetter
        >>>
        >>> from hbutils.string import format_tree
        >>>
        >>> tree = (
        >>>     'foo', [
        >>>         ('bar', [
        >>>             ('a', []),
        >>>             ('b', []),
        >>>         ]),
        >>>         ('baz', []),
        >>>         ('qux', [
        >>>             ('c\nd', []),
        >>>         ]),
        >>>     ],
        >>> )
        >>> format_tree(tree, format_node=itemgetter(0), get_children=itemgetter(1))
        foo
        ├── bar
        │   ├── a
        │   └── b
        ├── baz
        └── qux
            └── c
                d
    """
    if 'ASCII' in (encoding or _DEFAULT_ENCODING).upper():
        _chars = _ASCII_CHARS
    else:
        _chars = _UTF8_CHARS

    return os.linesep.join(itertools.chain(
        [format_node(node)],
        _format_tree(node, format_node, get_children, u'', _chars),
        [u''],
    ))



import platform
from enum import IntEnum, unique
from functools import lru_cache

from ...model import int_enum_loads

__all__ = [
    'is_linux', 'is_windows', 'is_macos', 'is_darwin'
]


@int_enum_loads(name_preprocess=str.upper)
@unique
class OSType(IntEnum):
    LINUX = 1
    WINDOWS = 2
    DARWIN = 3
    JAVA = 4


@lru_cache()
def _get_os_type() -> OSType:
    return OSType.loads(platform.system())


def is_linux() -> bool:
    """
    Overview:
        Return ``True`` if current operating system is linux, otherwise return ``False``.

    :return: This OS is linux or not.
    """
    return _get_os_type() == OSType.LINUX


def is_windows() -> bool:
    """
    Overview:
        Return ``True`` if current operating system is windows, otherwise return ``False``.

    :return: This OS is windows or not.
    """
    return _get_os_type() == OSType.WINDOWS


def is_darwin() -> bool:
    """
    Overview:
        Return ``True`` if current operating system is darwin, otherwise return ``False``.

    :return: This OS is darwin or not.

    .. note::
        Darwin is macos.
    """
    return _get_os_type() == OSType.DARWIN


is_macos = is_darwin
"""
Alias of :func:`is_darwin`.
"""



"""
Overview:
    Collection module, include the basic utilities with collections like dict, list and tuple.
"""
from .dimension import *
from .functional import *
from .recover import *
from .sequence import *
from .stacked import *
from .structural import *



import ctypes
from typing import Dict, BinaryIO, List

from .base import CRangedIntType

__all__ = [
    'CUnsignedIntType',
    'c_uint8', 'c_uint16', 'c_uint32', 'c_uint64',
    'c_ubyte', 'c_ushort', 'c_uint', 'c_ulong', 'c_ulonglong',
]


class CUnsignedIntType(CRangedIntType):
    """
    Overview:
        Unsigned int type.
    """

    def __init__(self, size: int):
        """
        Constructor of :class:`CUnsignedIntType`.

        :param size: Size of the type.
        """
        self.__size = size
        CRangedIntType.__init__(
            self, self.__size,
            0, (1 << (8 * self.__size)) - 1,
        )

    def read(self, file: BinaryIO) -> int:
        """
        Read unsigned int value.

        :param file: Binary file, ``io.BytesIO`` is supported as well.
        :return: Unsigned int value.
        """
        result = 0
        for i, byte_ in enumerate(file.read(self.__size)):
            result |= byte_ << (8 * i)
        return result

    def write(self, file: BinaryIO, val: int):
        """
        Write unsigned int value to binary IO object.

        :param file: Binary file, ``io.BytesIO`` is supported as well.
        :param val: Unsigned int value to write.
        """
        if not isinstance(val, int):
            raise TypeError(f'Int value expected, but {repr(val)} found.')
        elif not (self.minimum <= val <= self.maximum):
            raise ValueError(f'Signed int value within '
                             f'[{self.minimum}, {self.maximum}] expected, but {repr(val)} found.')

        _bytes = []
        for _ in range(self.__size):
            _bytes.append(val & 0xff)
            val >>= 8

        file.write(bytes(_bytes))


c_uint8 = CUnsignedIntType(ctypes.sizeof(ctypes.c_uint8))
"""
Overview:
    Reading and writing unsigned integer with 8-bits.
    
Examples::
    >>> import io
    >>> from hbutils.binary import c_uint8
    >>> 
    >>> with io.BytesIO(b'\xde\xad\xbe\xef') as file:
    ...     print(c_uint8.read(file))
    ...     print(c_uint8.read(file))
    ...     print(c_uint8.read(file))
    ...     print(c_uint8.read(file))
    222
    173
    190
    239
    >>> with io.BytesIO() as file:
    ...     c_uint8.write(file, 222)
    ...     c_uint8.write(file, 173)
    ...     c_uint8.write(file, 190)
    ...     c_uint8.write(file, 239)
    ...     print(file.getvalue())
    b'\xde\xad\xbe\xef'
"""
c_uint16 = CUnsignedIntType(ctypes.sizeof(ctypes.c_uint16))
"""
Overview:
    Reading and writing unsigned integer with 16-bits.
    
Examples::
    >>> import io
    >>> from hbutils.binary import c_uint16
    >>> 
    >>> with io.BytesIO(b'\xde\xad\xbe\xef\x12\x34\x56\x78') as file:
    ...     print(c_uint16.read(file))
    ...     print(c_uint16.read(file))
    ...     print(c_uint16.read(file))
    ...     print(c_uint16.read(file))
    44510
    61374
    13330
    30806
    >>> with io.BytesIO() as file:
    ...     c_uint16.write(file, 44510)
    ...     c_uint16.write(file, 61374)
    ...     c_uint16.write(file, 13330)
    ...     c_uint16.write(file, 30806)
    ...     print(file.getvalue())
    b'\xde\xad\xbe\xef\x124Vx'
"""
c_uint32 = CUnsignedIntType(ctypes.sizeof(ctypes.c_uint32))
"""
Overview:
    Reading and writing unsigned integer with 32-bits.
    
Examples::
    >>> import io
    >>> from hbutils.binary import c_uint32
    >>> 
    >>> with io.BytesIO(b'\xde\xad\xbe\xef\x12\x34\x56\x78') as file:
    ...     print(c_uint32.read(file))
    ...     print(c_uint32.read(file))
    4022250974
    2018915346
    >>> with io.BytesIO() as file:
    ...     c_uint32.write(file, 4022250974)
    ...     c_uint32.write(file, 2018915346)
    ...     print(file.getvalue())
    b'\xde\xad\xbe\xef\x124Vx'
"""
c_uint64 = CUnsignedIntType(ctypes.sizeof(ctypes.c_uint64))
"""
Overview:
    Reading and writing unsigned integer with 64-bits.
    
Examples::
    >>> import io
    >>> from hbutils.binary import c_uint64
    >>> 
    >>> with io.BytesIO(b'\xde\xad\xbe\xef\x12\x34\x56\x78') as file:
    ...     print(c_uint64.read(file))
    8671175388484775390
    >>> with io.BytesIO() as file:
    ...     c_uint64.write(file, 0x78563412efbeadde)
    ...     print(file.getvalue())
    b'\xde\xad\xbe\xef\x124Vx'
"""

_EXIST_TYPES: List[CUnsignedIntType] = [
    c_uint8,
    c_uint16,
    c_uint32,
    c_uint64,
]
_SIZE_TO_INT_TYPE: Dict[int, CUnsignedIntType] = {
    item.size: item
    for item in _EXIST_TYPES
}


def _get_from_raw(tp) -> CUnsignedIntType:
    return _SIZE_TO_INT_TYPE[ctypes.sizeof(tp)]


c_ubyte = _get_from_raw(ctypes.c_ubyte)
"""
Alias for :data:`c_uint8`.
"""
c_ushort = _get_from_raw(ctypes.c_ushort)
"""
Alias for :data:`c_uint16`.
"""
c_uint = _get_from_raw(ctypes.c_uint)
"""
Alias for :data:`c_uint32` (in 64-bits OS).

.. note::
    Size of :data:`c_uint` is the same as that in C language, which mainly depends on CPU and OS.
"""
c_ulong = _get_from_raw(ctypes.c_ulong)
"""
Alias for :data:`c_uint64` (in 64-bits OS).

.. note::
    Size of :data:`c_ulong` is the same as that in C language, which mainly depends on CPU and OS.
"""
c_ulonglong = _get_from_raw(ctypes.c_ulonglong)
"""
Alias for :data:`c_uint64` (in 64-bits OS).

.. note::
    Size of :data:`c_ulonglong` is the same as that in C language, which mainly depends on CPU and OS.
"""



from abc import ABCMeta

from .version import VersionInfo
from ...system import python_version, package_version, is_windows, is_linux, is_darwin, is_cpython, is_pypy, \
    is_ironpython, is_jython

__all__ = [
    'vpython', 'vpip',
    'OS', 'Impl',
]

vpython = VersionInfo(lambda: python_version())
vpython.__doc__ = """
Overview:
    Python version expression.

Examples::
    >>> import platform
    >>> import unittest
    >>> from hbutils.testing import vpython
    >>> 
    >>> class TestMyCase(unittest.TestCase):  # on python 3.6
    ...     def test_anytime(self):
    ...         assert 2 + 1 == 3
    ...
    ...     @unittest.skipUnless('3.6' <= vpython < '3.7', 'py36 only')  # will run
    ...     def test_on_python36(self):
    ...         assert platform.python_version_tuple()[:2] == ('3', '6')
    ...
    ...     @unittest.skipUnless('3.7' <= vpython < '3.8', 'py37 only')  # will skip
    ...     def test_on_python37(self):
    ...         assert platform.python_version_tuple()[:2] == ('3', '7')
    ...
    >>> unittest.main()
    ..s
    ----------------------------------------------------------------------
    Ran 3 tests in 0.000s
    OK (skipped=1)
"""


class PipVersionInfo(VersionInfo):
    def __call__(self, name: str) -> VersionInfo:
        return VersionInfo(lambda: package_version(name))


vpip = PipVersionInfo(lambda: package_version('pip'))
vpip.__doc__ = """
Overview:
    Pip version expression

Examples::
    >>> import unittest
    >>> from hbutils.testing import vpip
    >>> 
    >>> class TestMyCase(unittest.TestCase):
    ...     def test_1_anytime(self):
    ...         assert 2 + 1 == 3
    ...
    ...     @unittest.skipUnless(vpip >= '21', 'pip21+ only')  # pip>=21
    ...     def test_2_on_pip21plus(self):
    ...         assert True
    ...
    ...     @unittest.skipUnless(vpip('pip') >= '21', 'pip21+ only')  # the same as above
    ...     def test_3_on_pip21plus2(self):
    ...         assert True
    ...
    ...     @unittest.skipUnless(vpip('setuptools') >= '45' or vpip('build') >= '0.8', '')  # setuptools>=45 or build>=0.8
    ...     def test_4_on_setuptools_or_build(self):
    ...         assert True
    ...
    ...     @unittest.skipUnless(not vpip and vpip('build') >= '0.8', '')  # pip not installed, and build>=0.8
    ...     def test_5_on_nopip_and_build(self):
    ...         assert True
    ... 
    >>> unittest.main()
    ....s
    ----------------------------------------------------------------------
    Ran 5 tests in 0.000s
    OK (skipped=1)
"""


class OS(metaclass=ABCMeta):
    """
    Overview:
        Expressions for operating system.

    Examples::
        >>> import unittest
        >>> from hbutils.testing import OS
        >>>
        >>> class TestMyCase(unittest.TestCase):  # on Linux
        ...     def test_1_anytime(self):
        ...         assert 2 + 1 == 3
        ...
        ...     @unittest.skipUnless(OS.linux, 'linux only')  # only run on Linux
        ...     def test_2_linux(self):
        ...         assert True
        ...
        ...     @unittest.skipUnless(OS.windows, 'windows only')  # only run on Windows
        ...     def test_2_windows(self):
        ...         assert True
        ...
        ...     @unittest.skipUnless(OS.macos, 'macos only')  # only run on macOS
        ...     def test_4_macos(self):
        ...         assert True
        ...
        >>> unittest.main()
        ..ss
        ----------------------------------------------------------------------
        Ran 4 tests in 0.001s
        OK (skipped=2)
    """
    windows = is_windows()
    """
    Is windows system or not, related to your local OS.
    """

    linux = is_linux()
    """
    Is linux system or not, related to your local OS.
    """

    darwin = is_darwin()
    """
    Is darwin (macos) system or not, related to your local OS.
    """
    macos = darwin
    """
    Alias for ``OS.darwin``.
    """


class Impl:
    """
    Overview:
        Expression for python implementation.
        See `platform.python_implementation() \
        <https://docs.python.org/3/library/platform.html#platform.python_implementation>`_ .

    Examples::
        >>> import unittest

from hbutils.testing import Impl
        >>>
        >>> class TestMyCase(unittest.TestCase):  # on CPython
        ...     def test_1_anytime(self):
        ...         assert 2 + 1 == 3
        ...
        ...     @unittest.skipUnless(Impl.cpython, 'cpython only')  # only run on CPython
        ...     def test_2_cpython(self):
        ...         assert True
        ...
        ...     @unittest.skipUnless(Impl.pypy, 'pypy only')  # only run on PyPy
        ...     def test_3_pypy(self):
        ...         assert True
        ...
        ...     @unittest.skipUnless(Impl.iron_python, 'ironpython only')  # only run on IronPython
        ...     def test_4_ironpython(self):
        ...         assert True
        ...
        ...     @unittest.skipUnless(Impl.jython, 'jython only')  # only run on Jython
        ...     def test_5_jython(self):
        ...         assert True
        ...
        >>> unittest.main()
        ..sss
        ----------------------------------------------------------------------
        Ran 5 tests in 0.000s
        OK (skipped=3)
    """
    cpython = is_cpython()
    """
    Is CPython (most-frequently-used python) or not, related to your local python.
    """

    iron_python = is_ironpython()
    """
    Is IronPython or not, related to your local python.
    """

    jython = is_jython()
    """
    Is Jython (java-based python) or not, related to your local python.
    """

    pypy = is_pypy()
    """
    Is PyPy or not, related to your local python.
    """



"""
Overview:
    Implementation of observer pattern.
    See `Observer Pattern - Wikipedia <https://en.wikipedia.org/wiki/Observer_pattern>`_.
"""
from enum import Enum
from typing import Union, Type, TypeVar, Dict, Callable, Tuple, Any, List

__all__ = [
    'Observable'
]

_EventSetType = Union[Type[Enum], list, tuple]


def _auto_members(events: _EventSetType):
    if isinstance(events, type) and issubclass(events, Enum):
        return list(events.__members__.values())
    elif isinstance(events, (list, tuple)):
        return list(events)
    else:
        raise TypeError(f'Invalid event set - {repr(events)}.')


def _get_object_id(obj) -> Tuple[str, int]:
    try:
        return 'hash', hash(obj)
    except TypeError:
        return 'id', id(obj)


_EventType = TypeVar('_EventType')
_SubscriberType = TypeVar('_SubscriberType')
_CallbackType = TypeVar('_CallbackType', bound=Callable[..., Any])


class _CallbackWrapper:
    def __init__(self, callback):
        from ..reflection import dynamic_call, sigsupply
        self.raw = callback
        self._callback = dynamic_call(sigsupply(callback, lambda x: None))

    def __call__(self, *args, **kwargs):
        return self._callback(*args, **kwargs)


class Observable:
    """
    Overview:
        Observable object.

        * :meth:`subscribe` can be used for subscribing on specific event.
        * :meth:`unsubscribe` can be used for unsubscribing from specific event.
        * :meth:`dispatch` can be used for broadcasting a specific event, and all the subscribed callback will be \
            triggered.

    Examples::
        >>> from enum import IntEnum, unique
        >>> from hbutils.design import Observable
        >>>
        >>> @unique
        ... class MyIntEnum(IntEnum):
        ...     A = 1
        ...     B = 2
        >>>
        >>> o = Observable(MyIntEnum)
        >>> list_a, list_b = [], []
        >>> o.subscribe(MyIntEnum.A, list_a, 'append')  # use list_a.append
        >>> o.subscribe(MyIntEnum.B, list_a, lambda v: list_a.append(v))  # custom function.
        >>> o.subscribe(MyIntEnum.A, list_b, 'append')  # use list_b.append
        >>>
        >>> list_a, list_b
        ([], [])
        >>> o.dispatch(MyIntEnum.A)
        >>> list_a, list_b
        ([<MyIntEnum.A: 1>], [<MyIntEnum.A: 1>])
        >>> o.dispatch(MyIntEnum.B)
        >>> list_a, list_b
        ([<MyIntEnum.A: 1>, <MyIntEnum.B: 2>], [<MyIntEnum.A: 1>])
        >>>
        >>> o.unsubscribe(MyIntEnum.A, list_a)
        >>> o.dispatch(MyIntEnum.A)
        >>> list_a, list_b
        ([<MyIntEnum.A: 1>, <MyIntEnum.B: 2>], [<MyIntEnum.A: 1>, <MyIntEnum.A: 1>])
        >>> o.dispatch(MyIntEnum.B)
        >>> list_a, list_b
        ([<MyIntEnum.A: 1>, <MyIntEnum.B: 2>, <MyIntEnum.B: 2>], [<MyIntEnum.A: 1>, <MyIntEnum.A: 1>])
    """

    def __init__(self, events: _EventSetType):
        """
        Constructor of :class:`Observable`.

        :param events: Set of events, can be a list, tuple or an enum class.

        .. note::
            When enum is used, its values will be used as events. For example:

            >>> from enum import IntEnum
            >>> from hbutils.design import Observable
            >>>
            >>> class MyIntEnum(IntEnum):
            ...     A = 1
            ...     B = 2
            >>>
            >>> # equals to `Observable([MyIntEnum.A, MyIntEnum.B])`
            ... o = Observable(MyIntEnum)
            >>> o._events  # just for explanation, do not do this on actual use
            {<MyIntEnum.A: 1>: {}, <MyIntEnum.B: 2>: {}}
        """
        events = _auto_members(events)
        self._observers: Dict[Tuple[str, int], _SubscriberType] = {}
        self._events: Dict[_EventType, Dict[Tuple[str, int], _CallbackWrapper]] = {e: {} for e in events}

    def subscribers(self, event: _EventType) -> List[_SubscriberType]:
        """
        Get subscribers of the given ``event``.

        :param event: Event for querying.
        :return: A list of subscribers.
        """
        return [self._observers[id_] for id_ in self._get_subscriptions(event).keys()]

    def subscriptions(self, event: _EventType) -> List[Tuple[_SubscriberType, _CallbackType]]:
        """
        Get subscriptions of the given ``event``.

        :param event: Event for querying.
        :return: A list of tuples with subscribers and their callbacks.
        """
        return [
            (self._observers[id_], callback.raw)
            for id_, callback in self._get_subscriptions(event).items()
        ]

    def _get_subscriptions(self, event: _EventType) -> Dict[Tuple[str, int], _CallbackWrapper]:
        return self._events[event]

    def _put_subscription(self, event: _EventType, subscriber: _SubscriberType, callback: _CallbackType):
        subscriber_id = _get_object_id(subscriber)
        self._get_subscriptions(event)[subscriber_id] = _CallbackWrapper(callback)
        self._observers[subscriber_id] = subscriber

    def _del_subscription(self, event: _EventType, subscriber: _SubscriberType):
        try:
            del self._get_subscriptions(event)[_get_object_id(subscriber)]
        except KeyError:
            raise KeyError(subscriber)

    def subscribe(self, event: _EventType, subscriber: _SubscriberType,
                  callback: Union[_CallbackType, str, None] = None):
        """
        Subscribe to the given ``event``.

        :param event: Event to be subscribed.
        :param subscriber: Subscriber of this subscription.
        :param callback: Callback function. If ``str`` is given, method with this name on ``subscriber`` will be used. \
            Default is ``None`` which means the ``update`` method on ``subscriber`` will be used.

        .. note::
            Callback function should have no more than 2 positional arguments. For example:

            >>> o.subscribe(MyIntEnum.A, 'user1', lambda: 2)  # ok
            >>> o.subscribe(MyIntEnum.A, 'user2', lambda event: 2)  # ok
            >>> o.subscribe(MyIntEnum.A, 'user3', lambda event, obs: 2)  # ok
            >>> o.subscribe(MyIntEnum.A, 'user4', lambda x, y, z: 2)  # X
        """
        if callback is None:
            callback = getattr(subscriber, 'update')
        elif isinstance(callback, str):
            callback = getattr(subscriber, callback)
        if not callable(callback):
            raise TypeError(f'Callback should be callable, but {repr(callback)} found.')

        self._put_subscription(event, subscriber, callback)

    def unsubscribe(self, event: _EventType, subscriber: _SubscriberType):
        """
        Unsubscribe from the given ``event``.

        :param event: Event to be unsubscribed.
        :param subscriber: Subscriber of this unsubscription.
        """
        self._del_subscription(event, subscriber)

    def dispatch(self, event: _EventType):
        """
        Dispatch event.

        :param event: Event to be dispatched.
        """
        for _, callback in self._get_subscriptions(event).items():
            callback(event, self)



import inspect
from contextlib import contextmanager
from dataclasses import dataclass
from functools import wraps
from itertools import chain
from typing import Iterator, Tuple, Any, Union, List, Dict
from unittest.mock import patch, MagicMock

from hbutils.reflection import quick_import_object, nested_with

__all__ = [
    'isolated_entry_points',
]


@dataclass
class _FakeEntryPoint:
    name: str
    group: str
    dist: object

    def load(self):
        return self.dist


_max_fake_id = 0


def _fake_id() -> int:
    global _max_fake_id
    _max_fake_id += 1
    return _max_fake_id


def _fake_entry_name() -> str:
    return f'unnamed_fake_entry_{_fake_id()}'


def _yield_from_units(fes, auto_import: bool = True) -> Iterator[Tuple[str, Any]]:
    if isinstance(fes, (list, tuple)):
        for item in fes:
            if isinstance(item, tuple) and len(item) == 2:
                name, dist = item
                if auto_import and isinstance(dist, str):
                    dist, _, _ = quick_import_object(dist)
            elif auto_import and isinstance(item, str):
                dist, _, name = quick_import_object(item)
            elif hasattr(item, '__name__'):
                name, dist = item.__name__, item
            else:
                name, dist = _fake_entry_name(), item

            yield name, dist

    elif isinstance(fes, dict):
        for name, dist in fes.items():
            if auto_import and isinstance(dist, str):
                dist, _, _ = quick_import_object(dist)

            yield name, dist

    else:
        raise TypeError(f'Unknown type of fake entries - {fes!r}.')  # pragma: no cover


def _yield_fake_entries(group, fes, auto_import: bool = True) -> Iterator[_FakeEntryPoint]:
    for name, dist in _yield_from_units(fes, auto_import):
        yield _FakeEntryPoint(name, group, dist)


@contextmanager
def isolated_entry_points(group: str, fakes: Union[List, Dict[str, Any], None] = None,
                          auto_import: bool = True, clear: bool = False):
    """
    Overview:
        Isolation for :func:`pkg_resources.iter_entry_points` function.
        Can be used to fake the plugins, or just disable the installed plugins.

        ``importlib.metadata`` and ``importlib_metadata`` are supported now.

    :param group: Group name.
    :param fakes: Fake entry points. Dict or list are accepted.
    :param auto_import: Auto import the object from string. Default is ``True`` which means if \
        a string is given, dynamic import will be performed.
    :param clear: Clear the original entry points or not. Default is ``False`` which means the original \
        entry points will be kept and be able to be iterated.

    Examples::
        >>> import importlib.metadata
        >>> import importlib_metadata  # backport for py3.7
        >>> import pkg_resources  # deprecated
        >>>
        >>> from hbutils.testing import isolated_entry_points
        >>>
        >>> # nothing at the beginning
        >>> print({ep.name: ep.load() for ep in
        ...        pkg_resources.iter_entry_points('my_plugin')})
        {}
        >>> print(importlib.metadata.entry_points().get('my_plugin', None))
        None
        >>> print(importlib_metadata.entry_points(group='my_plugin'))
        ()
        >>>
        >>> # mock the plugins
        >>> with isolated_entry_points('my_plugin', [
        ...     (
        ...             'quick_import_object',  # name import
        ...             'hbutils.reflection.quick_import_object'
        ...     ),
        ...     ('func_filter', filter),  # named entry object
        ...     map,  # simple function
        ...     'hbutils.system.is_binary_file',  # simple import
        ... ]):
        ...     print({ep.name: ep.load() for ep in
        ...            pkg_resources.iter_entry_points('my_plugin')})
        ...     print(importlib.metadata.entry_points()['my_plugin'])
        ...     print(importlib_metadata.entry_points(group='my_plugin'))
        ...
        {'quick_import_object': <function quick_import_object at 0x7f6d0dd5ad40>, 'func_filter': <class 'filter'>, 'map': <class 'map'>, 'is_binary_file': <function is_binary_file at 0x7f6d0dcd8550>}
        [_FakeEntryPoint(name='quick_import_object', group='my_plugin', dist=<function quick_import_object at 0x7f6d0dd5ad40>), _FakeEntryPoint(name='func_filter', group='my_plugin', dist=<class 'filter'>), _Fak
        eEntryPoint(name='map', group='my_plugin', dist=<class 'map'>), _FakeEntryPoint(name='is_binary_file', group='my_plugin', dist=<function is_binary_file at 0x7f6d0dcd8550>)]
        [_FakeEntryPoint(name='quick_import_object', group='my_plugin', dist=<function quick_import_object at 0x7f6d0dd5ad40>), _FakeEntryPoint(name='func_filter', group='my_plugin', dist=<class 'filter'>), _Fak
        eEntryPoint(name='map', group='my_plugin', dist=<class 'map'>), _FakeEntryPoint(name='is_binary_file', group='my_plugin', dist=<function is_binary_file at 0x7f6d0dcd8550>)]
        >>> with isolated_entry_points('my_plugin', {
        ...     'func_map': map,
        ...     'func_binary': 'hbutils.system.is_binary_file'
        ... }):
        ...     print({ep.name: ep.load() for ep in
        ...            pkg_resources.iter_entry_points('my_plugin')})
        ...     print(importlib.metadata.entry_points()['my_plugin'])
        ...     print(importlib_metadata.entry_points(group='my_plugin'))
        ...
        {'func_map': <class 'map'>, 'func_binary': <function is_binary_file at 0x7f6d0dcd8550>}
        [_FakeEntryPoint(name='func_map', group='my_plugin', dist=<class 'map'>), _FakeEntryPoint(name='func_binary', group='my_plugin', dist=<function is_binary_file at 0x7f6d0dcd8550>)]
        [_FakeEntryPoint(name='func_map', group='my_plugin', dist=<class 'map'>), _FakeEntryPoint(name='func_binary', group='my_plugin', dist=<function is_binary_file at 0x7f6d0dcd8550>)]
        >>> # nothing at the ending
        >>> print({ep.name: ep.load() for ep in
        ...        pkg_resources.iter_entry_points('my_plugin')})
        {}
        >>> print(importlib.metadata.entry_points().get('my_plugin', None))
        None
        >>> print(importlib_metadata.entry_points(group='my_plugin'))
        ()

        .. warning::
            The ``pkg_resources`` package is no longer officially supported.
            However, certain libraries that rely on hbutils are still in use,
            hence temporary support will be provided. The official guidance must be followed to
            migrate to ``importlib.metadata`` at the earliest opportunity.
            In addition, support for the ``pkg_resources`` package in this function will be
            discontinued in the next major version.
    """
    if fakes is not None and not isinstance(fakes, (list, tuple, dict)):
        raise TypeError(f'Unknown entry point type - {fakes!r}.')

    from pkg_resources import iter_entry_points as _origin_iep
    group_name = group

    # noinspection PyShadowingNames
    @wraps(_origin_iep)
    def _new_iter_func(group, name=None):
        _exist_names = set()

        def _check_name(x) -> bool:
            if (name is None or x.name == name) and x.name not in _exist_names:
                _exist_names.add(x.name)
                return True
            else:
                return False

        if group == group_name:
            mocked = _yield_fake_entries(group_name, fakes or [], auto_import)
            if not clear:
                mocked = chain(mocked, _origin_iep(group, name))
            yield from filter(_check_name, mocked)
        else:
            yield from _origin_iep(group, name)

    try:
        import importlib_metadata as _py37_metadata
    except (ModuleNotFoundError, ImportError):
        _py37_metadata = None
    else:
        _py37_origin_entry_points = _py37_metadata.entry_points

        @wraps(_py37_origin_entry_points)
        def _py37_entry_points(**kwargs):
            kwargs = {key: value for key, value in kwargs.items() if value}
            group_ = kwargs.get('group', None)
            name = kwargs.get('name', None)
            _exist_names = set()

            def _check_name(x) -> bool:
                if ((isinstance(x, _py37_metadata.EntryPoint) and x.matches(**kwargs)) or
                    (not isinstance(x, _py37_metadata.EntryPoint) and (name is None or x.name == name))) and \
                        x.name not in _exist_names:
                    _exist_names.add(x.name)
                    return True
                else:
                    return False

            if group_ is None or group_ == group_name:
                mocked = _yield_fake_entries(group_name, fakes or [], auto_import)
                if not clear:
                    mocked = chain(mocked, _py37_origin_entry_points(**kwargs))
                # noinspection PyTypeChecker
                return list(filter(_check_name, mocked))
            else:
                return list(_py37_origin_entry_points(**kwargs))

    try:
        import importlib.metadata as _py38_metadata
    except (ModuleNotFoundError, ImportError):
        _py38_metadata = None
    else:
        _py38_origin_entry_points = _py38_metadata.entry_points
        _py38_func_has_params = bool(inspect.signature(_py38_metadata.entry_points).parameters)

        @wraps(_py38_origin_entry_points)
        def _py38_entry_points(**kwargs):
            kwargs = {key: value for key, value in kwargs.items() if value}
            group_ = kwargs.get('group', None)
            name = kwargs.get('name', None)
            _exist_names = set()

            def _check_name(x) -> bool:
                if ((isinstance(x, _py38_metadata.EntryPoint) and x.matches(**kwargs)) or
                    (not isinstance(x, _py38_metadata.EntryPoint) and (name is None or x.name == name))) and \
                        x.name not in _exist_names:
                    _exist_names.add(x.name)
                    return True
                else:
                    return False

            # noinspection PyArgumentList
            _base_result = _py38_origin_entry_points(**kwargs)
            if isinstance(_base_result, dict):  # kwargs must be empty
                _retval = _base_result.copy()
                mocked = _yield_fake_entries(group_name, fakes or [], auto_import)
                if not clear:
                    mocked = chain(mocked, (_retval.get(group_name, None) or []))

                # noinspection PyTypeChecker
                _retval[group_name] = (list if _py38_func_has_params else tuple)(filter(_check_name, mocked))
                return _retval
            else:
                if group_ is None or group_ == group_name:
                    mocked = _yield_fake_entries(group_name, fakes or [], auto_import)
                    if not clear:
                        mocked = chain(mocked, _base_result)

                    # noinspection PyTypeChecker
                    return list(filter(_check_name, mocked))
                else:

                    return _base_result

    mocks = []
    mocks.append(patch('pkg_resources.iter_entry_points', MagicMock(side_effect=_new_iter_func)))
    if _py37_metadata:
        mocks.append(patch('importlib_metadata.entry_points', MagicMock(side_effect=_py37_entry_points)))
    if _py38_metadata:
        mocks.append(patch('importlib.metadata.entry_points', MagicMock(side_effect=_py38_entry_points)))

    with nested_with(*mocks):
        yield



"""
Overview:
    Functions to deal with encoding binary data easily.
"""
import sys
from typing import Optional, List

import chardet

from ..collection import unique

_DEFAULT_ENCODING = 'utf-8'
_DEFAULT_PREFERRED_ENCODINGS = ['utf-8', 'gbk', 'gb2312', 'gb18030', 'big5']  # common encodings for chinese

__all__ = [
    'auto_decode'
]


def _decode(data: bytes, encoding: str) -> str:
    return data.decode(encoding)


def auto_decode(data: bytes, encoding: Optional[str] = None, prefers: Optional[List[str]] = None) -> str:
    r"""
    Overview:
        Auto decode binary data to string, the encoding mode will be automatically detected.

    Arguments:
        - data (:obj:`bytes`): Original binary data to be decoded.
        - encoding (:obj:`Optional[str]`): Encoding mode to be used, default is ``None`` which \
            means this function need to automatically detect the encoding.
        - prefers (:obj:`Optional[List[str]]`): Prefered encodings.

    Returns:
        - str (:obj:`str`): Decoded string.

    Examples::

        >>> auto_decode(b'kdsfjldsjflkdsmgds')  # 'kdsfjldsjflkdsmgds'
        >>> auto_decode(b'\xd0\x94\xd0\xbe\xd0\xb1\xd1\x80\xd1\x8b\xd0\xb9 \xd0'
        ...             b'\xb2\xd0\xb5\xd1\x87\xd0\xb5\xd1\x80')  # "Добрый вечер"
        >>> auto_decode(b'\xa4\xb3\xa4\xf3\xa4\xd0\xa4\xf3\xa4\xcf')  # "こんばんは"
        >>> auto_decode(b'\xcd\xed\xc9\xcf\xba\xc3')  # "晚上好"
    """
    if encoding:
        return _decode(data, encoding)
    else:
        if prefers is None:
            prefers = _DEFAULT_PREFERRED_ENCODINGS
        _elist = filter(bool, unique([
            *prefers,
            sys.getdefaultencoding(),
            chardet.detect(data)['encoding']
        ]))

        last_err = None
        for enc in _elist:
            try:
                return _decode(data, enc)
            except UnicodeDecodeError as err:
                if last_err is None or err.start > last_err.start:
                    last_err = err

        raise last_err



"""
Overview:
    Utilities for building context variables on thread level.

    This is useful when implementing a with-block-based syntax.
    For example:

        >>> from contextlib import contextmanager
        >>> from hbutils.reflection import context
        >>>
        >>> # developer's view
        ... @contextmanager
        ... def use_mul():  # set 'mul' to `True` in its with-block
        ...     with context().vars(mul=True):
        ...         yield
        >>>
        >>> def calc(a, b):  # logic of `calc` will be changed when 'mul' is given
        ...     if context().get('mul', None):
        ...         return a * b
        ...     else:
        ...         return a + b
        >>>
        >>> # user's view (magic-liked, isn't it?)
        ... print(calc(3, 5))  # 3 + 5
        8
        >>> with use_mul():
        ...     print(calc(3, 5))  # changed to 3 * 5
        15
        >>> print(calc(3, 5))  # back to 3 + 5, again :)
        8
"""
import collections.abc

from contextlib import contextmanager
from functools import wraps
from multiprocessing import current_process
from threading import current_thread
from typing import Tuple, TypeVar, Iterator, Mapping, Optional, ContextManager, Any

__all__ = [
    'context', 'cwrap',
    'nested_with', 'conditional_with',
]


def _get_pid() -> int:
    return current_process().pid


def _get_tid() -> int:
    return current_thread().ident


def _get_context_id() -> Tuple[int, int]:
    return _get_pid(), _get_tid()


_global_contexts = {}

_KeyType = TypeVar('_KeyType', bound=str)
_ValueType = TypeVar('_ValueType')


class ContextVars(collections.abc.Mapping):
    """
    Overview:
        Context variable management.

    .. note::
        This class is inherited from :class:`collections.abc.Mapping`.
        Main features of mapping object (such as ``__getitem__``, ``__len__``, ``__iter__``) are supported.
        See `Collections Abstract Base Classes \
        <https://docs.python.org/3/library/collections.abc.html#collections-abstract-base-classes>`_.

    .. warning::
        This object should be singleton on thread level.
        It is not recommended constructing manually.
    """

    def __init__(self, **kwargs):
        """
        Constructor of :class:`ContextVars`.

        :param kwargs: Initial context values.
        """
        self._vars = dict(kwargs)

    @contextmanager
    def _with_vars(self, params: Mapping[_KeyType, _ValueType], clear: bool = False):
        # initialize new values
        _origin = dict(self._vars)
        self._vars.update(params)
        if clear:
            for key in list(self._vars.keys()):
                if key not in params:
                    del self._vars[key]

        try:
            yield
        finally:
            # de-initialize, recover changed values
            for k in set(_origin.keys()) | set(self._vars.keys()):
                if k not in _origin:
                    del self._vars[k]
                else:
                    self._vars[k] = _origin[k]

    @contextmanager
    def vars(self, **kwargs):
        """
        Adding variables into context of ``with`` block.

        :param kwargs: Additional context variables.

        Examples::
            >>> from hbutils.reflection import context
            >>>
            >>> def var_detect():
            ...     if context().get('var', None):
            ...         print(f'Var detected, its value is {context()["var"]}.')
            ...     else:
            ...         print('Var not detected.')
            >>>
            >>> var_detect()
            Var not detected.
            >>> with context().vars(var=1):
            ...     var_detect()
            Var detected, its value is 1.
            >>> var_detect()
            Var not detected.

        .. note::
            See :func:`context`.

        """
        with self._with_vars(kwargs, clear=False):
            yield

    @contextmanager
    def inherit(self, context_: 'ContextVars'):
        """
        Inherit variables from the given context.

        :param context_: :class:`ContextVars` object to inherit from.

        .. note::
            After :meth:`inherit` is used, **the original variables which not present in the given ``context_`` \
            will be removed**. This is different from :meth:`vars`, so attention.
        """
        with self._with_vars(context_._vars, clear=True):
            yield

    def __getitem__(self, key: _KeyType):
        return self._vars[key]

    def __len__(self) -> int:
        return len(self._vars)

    def __iter__(self) -> Iterator[_KeyType]:
        return self._vars.__iter__()


def context() -> ContextVars:
    """
    Overview:
        Get context object in this thread.

    :return: Context object in this thread.

    .. note::
        This result is unique on one thread.
    """
    _context_id = _get_context_id()
    if _context_id not in _global_contexts:
        _context = ContextVars()
        _global_contexts[_context_id] = _context

    return _global_contexts[_context_id]


def cwrap(func, *, context_: Optional[ContextVars] = None, **vars_):
    """
    Overview:
        Context wrap for functions.

    :param func: Original function to wrap.
    :param context_: Context for inheriting. Default is ``None`` which means :func:`context`'s result will be used.
    :param vars_: External variables after inherit context.

    .. note::
        :func:`cwrap` is important when you need to pass the current context into thread.
        And **it is compitable on all platforms**.

        For example:

            >>> from threading import Thread
            >>> from hbutils.reflection import context, cwrap
            >>>
            >>> def var_detect():
            ...     if context().get('var', None):
            ...         print(f'Var detected, its value is {context()["var"]}.')
            ...     else:
            ...         print('Var not detected.')
            >>>
            >>> with context().vars(var=1):  # no inherit, vars will be lost in thread
            ...     t = Thread(target=var_detect)
            ...     t.start()
            ...     t.join()
            Var not detected.
            >>> with context().vars(var=1):  # with inherit, vars will be kept in thread
            ...     t = Thread(target=cwrap(var_detect))
            ...     t.start()
            ...     t.join()
            Var detected, its value is 1.

    .. warning::
        :func:`cwrap` **is not compitable on Windows or Python3.8+ on macOS** when creating **new process**.
        Please pass in direct arguments by ``args`` argument of :class:`Process`.
        If you insist on using :func:`context` feature, you need to pass the context object into the sub process.
        For example:

            >>> from contextlib import contextmanager
            >>> from multiprocessing import Process
            >>> from hbutils.reflection import context
            >>>
            >>> # developer's view
            ... @contextmanager
            ... def use_mul():  # set 'mul' to `True` in its with-block
            ...     with context().vars(mul=True):
            ...         yield
            >>>
            >>> def calc(a, b):  # logic of `calc` will be changed when 'mul' is given
            ...     if context().get('mul', None):
            ...         print(a * b)
            ...     else:
            ...         print(a + b)
            >>>
            >>> def _calc(a, b, ctx=None):
            ...     with context().inherit(ctx or context()):
            ...         return calc(a, b)
            >>>
            >>> # user's view
            ... if __name__ == '__main__':
            ...     calc(3, 5)  # 3 + 5
            ...     with use_mul():
            ...         p = Process(target=_calc, args=(3, 5, context()))
            ...         p.start()
            ...         p.join()
            ...     calc(3, 5)  # back to 3 + 5, again :)
            8
            15
            8

    """
    context_ = context_ or context()

    @wraps(func)
    def _new_func(*args, **kwargs):
        with context().inherit(context_):
            with context().vars(**vars_):
                return func(*args, **kwargs)

    return _new_func


def _yield_nested_for(contexts, depth, items):
    if depth >= len(contexts):
        yield tuple(items)
    else:
        with contexts[depth] as current_item:
            items.append(current_item)
            yield from _yield_nested_for(contexts, depth + 1, items)


@contextmanager
def nested_with(*contexts) -> ContextManager[Tuple[Any, ...]]:
    """
    Overview:
        Nested with, enter and exit multiple contexts.

    :param contexts: Contexts to manage.

    Examples::
        >>> import os.path
        >>> import pathlib
        >>> import tempfile
        >>> from contextlib import contextmanager
        >>> from hbutils.reflection import nested_with
        >>>
        >>> # allocate a temporary directory, and put one file inside
        >>> @contextmanager
        ... def opent(x):
        ...     with tempfile.TemporaryDirectory() as td:
        ...         pathlib.Path(os.path.join(td, f'{x}.txt')).write_text(f'this is {x}!')
        ...         yield td
        >>>
        >>> # let's try it
        >>> with opent(1) as d:
        ...     print(os.listdir(d))
        ...     print(pathlib.Path(f'{d}/1.txt').read_text())
        ['1.txt']
        this is 1!
        >>> # open 5 temporary directories at one time
        >>> with nested_with(*map(opent, range(5))) as ds:
        ...     for d in ds:
        ...         print(d)
        ...         print(os.path.exists(d), os.listdir(d))
        ...         print(pathlib.Path(f'{d}/{os.listdir(d)[0]}').read_text())
        /tmp/tmp3u1984br
        True ['0.txt']
        this is 0!
        /tmp/tmp0yx56hv0
        True ['1.txt']
        this is 1!
        /tmp/tmpu_33drm3
        True ['2.txt']
        this is 2!
        /tmp/tmpqal_vzgi
        True ['3.txt']
        this is 3!
        /tmp/tmpy99_wwtt
        True ['4.txt']
        this is 4!
        >>> # these directories are released now
        >>> for d in ds:
        ...     print(d)
        ...     print(os.path.exists(d))  # not exist anymore
        /tmp/tmp3u1984br
        False
        /tmp/tmp0yx56hv0
        False
        /tmp/tmpu_33drm3
        False
        /tmp/tmpqal_vzgi
        False
        /tmp/tmpy99_wwtt
        False
    """
    yield from _yield_nested_for(contexts, 0, [])


@contextmanager
def conditional_with(ctx, cond):
    """
    Overview:
        Conditional create context.

    :param ctx: Context object.
    :param cond: Condition for create or not.

    Examples::
        Here is an example of conditionally create a temporary directory.

        >>> import os.path
        >>>
        >>> from hbutils.reflection import conditional_with
        >>> from hbutils.system import TemporaryDirectory
        >>>
        >>> # create
        >>> with conditional_with(TemporaryDirectory(), cond=True) as td:
        ...     print('td:', td)
        ...     print('exist:', os.path.exists(td))
        ...     print('isdir:', os.path.isdir(td))
        ...
        td: /tmp/tmp07lpb9ah
        exist: True
        isdir: True
        >>> # not create
        >>> with conditional_with(TemporaryDirectory(), cond=False) as td:
        ...     print('td:', td)
        ...
        td: None
    """
    if cond:
        with ctx as f:
            yield f
    else:
        yield None



"""
Overview:
    Useful color utilities based on color model
"""
import math
import random
from typing import Iterator, Union, Sequence, Tuple, Callable

from .model import Color
from ..algorithm import linear_map

__all__ = [
    'visual_distance',
    'rnd_colors',
    'linear_gradient',
]


def _to_color(color: Color):
    if isinstance(color, Color):
        return color
    else:
        return Color(color)


def visual_distance(c1: Union[Color, str], c2: Union[Color, str]) -> float:
    """
    Overview:
        Get distance of 2 colors.

    Arguments:
        - c1 (:obj:`Color`): First color.
        - c2 (:obj:`Color`): Second color.

    Returns:
        - distance (:obj:`float`): Distance of the colors.

    Examples::
        >>> from hbutils.color import visual_distance, Color
        >>> visual_distance(
        ...     '#ff0000',
        ...     '#00ff00'
        ... )
        2.5495097567963922
        >>> visual_distance(
        ...     '#778800',
        ...     '#887700'
        ... )
        0.16996731711975946
    """
    c1, c2 = _to_color(c1), _to_color(c2)
    rgb1, rgb2 = c1.rgb, c2.rgb
    rmean = (rgb1.red + rgb2.red) / 2
    dr = rgb1.red - rgb2.red
    dg = rgb1.green - rgb2.green
    db = rgb1.blue - rgb2.blue

    return math.sqrt(
        (2 + rmean) * dr * dr +
        4 * dg * dg +
        (3 - rmean) * db * db
    )


def _dis_ratio(k):
    if k < 3:
        return 6.0
    if k < 6:
        return 2.0
    elif k < 8:
        return 0.7
    else:
        return 1.0


def rnd_colors(
        count, lightness=0.5, saturation=1.0, alpha=None,
        init_dis=4.0, lr=0.95, ur=1.5,
        rnd=None
) -> Iterator[Color]:
    """
    Overview:
        Generating random colors which are not similar.

    :param count: Count of colors.
    :param lightness: Lightness of the colors (in HLS color space), default is ``0.5``.
    :param saturation: Saturation of the colors (in HLS color space), default is ``1.0``.
    :param alpha: Alpha of the colors, default is ``None``.
    :param init_dis: Initial distance of colors, default is ``4.0``.
    :param lr: Lower ratio when generating, default is ``0.95``.
    :param ur: Upper ratio when generating, default is ``1.5``.
    :param rnd: Random object to be used, default is ``random.Random(0)``.

    Returns:
        - colors (:obj:`Iterator[Color]`): A iterator of colors.

    Examples::
        >>> from hbutils.color import rnd_colors
        >>> for c in rnd_colors(12):
        ...     print(c)
        #ff00ee
        #00ff00
        #009cff
        #ff006c
        #c9ff00
        #00f3ff
        #d100ff
        #ffaf00
        #00ff6c
        #4100ff
        #ff5300
        #46ff00

        >>> for c in rnd_colors(12, 0.8, 0.9):
        ...     print(c)
        #fa9ef4
        #9efaa1
        #9eb4fa
        #faa69e
        #c5fa9e
        #9ed6fa
        #f09efa
        #faf89e
        #9ef9fa
        #c09efa
        #fabe9e
        #9efaca
    """
    rnd = rnd or random.Random(0)
    min_distance = init_dis
    _exist_colors = []
    for i in range(count):
        try_cnt, total_try_cnt = 0, 0
        while True:
            new_color = Color.from_hls(rnd.random(), lightness, saturation)
            if not _exist_colors or all(
                    [visual_distance(color_, new_color) >= min_distance * _dis_ratio(i - j)
                     for j, color_ in enumerate(_exist_colors)]):
                _exist_colors.append(new_color)
                if total_try_cnt <= count * 2:
                    min_distance *= ur

                yield Color(new_color, alpha)
                break
            else:
                try_cnt += 1
                total_try_cnt += 1
                if try_cnt >= count * 2:
                    min_distance *= lr
                    try_cnt = 0


def linear_gradient(colors: Union[Sequence[Union[Color, str]], Sequence[Tuple[float, Union[Color, str]]]]) \
        -> Callable[[float], Color]:
    """
    Overview:
        Linear gradient of the colors.

    :param colors: Colors to gradient.
    :return: A mapping function for gradient.

    Examples::
        - Simple Linear Gradientation

        >>> from hbutils.color import linear_gradient
        >>>
        >>> f = linear_gradient(('red', 'yellow', 'lime'))
        >>> f(0)
        <Color red>
        >>> f(0.25)
        <Color #ff8000>
        >>> f(1 / 3)
        <Color #ffaa00>
        >>> f(0.5)
        <Color yellow>
        >>> f(2 / 3)
        <Color #aaff00>
        >>> f(0.75)
        <Color #80ff00>
        >>> f(1)
        <Color lime>

        - Complex Linear Gradientation

        >>> f = linear_gradient(((-0.2, 'red'), (0.7, '#ffff0044'), (1.1, 'lime')))
        >>> f(-0.2)
        <Color red, alpha: 1.000>
        >>> f(0)
        <Color #ff3900, alpha: 0.837>
        >>> f(0.25)
        <Color #ff8000, alpha: 0.633>
        >>> f(1 / 3)
        <Color #ff9700, alpha: 0.565>
        >>> f(0.5)
        <Color #ffc600, alpha: 0.430>
        >>> f(2 / 3)
        <Color #fff600, alpha: 0.294>
        >>> f(0.7)
        <Color yellow, alpha: 0.267>
        >>> f(0.75)
        <Color #dfff00, alpha: 0.358>
        >>> f(0.8)
        <Color #bfff00, alpha: 0.450>
        >>> f(1)
        <Color #40ff00, alpha: 0.817>
        >>> f(1.1)
        <Color lime, alpha: 1.000>

    """
    try:
        xys = [(x, _to_color(y)) for x, y in colors]
    except ValueError:
        pts = list(colors)
        n = len(pts)
        xys = [(i / (n - 1), _to_color(y)) for i, y in enumerate(colors)]

    rmap = linear_map([(x, y.rgb.red) for x, y in xys])
    gmap = linear_map([(x, y.rgb.green) for x, y in xys])
    bmap = linear_map([(x, y.rgb.blue) for x, y in xys])
    if any([y.alpha is not None for _, y in xys]):
        amap = linear_map([(x, y.alpha if y.alpha is not None else 1.0) for x, y in xys])
    else:
        # noinspection PyUnusedLocal
        def amap(x):
            return None

    def _gradient(x: float) -> Color:
        return Color((rmap(x), gmap(x), bmap(x)), amap(x))

    return _gradient



from types import GeneratorType
from typing import Tuple, Mapping, Optional, List, Iterator, Set


def _single_to_tuple(s: object) -> Tuple[object, ...]:
    if isinstance(s, (list, tuple, GeneratorType, range)):
        return tuple(s)
    else:
        return (s,)


def _single_dict_process(s: Mapping[str, object]) -> Mapping[str, Tuple[object, ...]]:
    return {key: _single_to_tuple(value) for key, value in s.items()}


def _check_keys(item: Mapping[str, object], names: Set[str]):
    for key in item.keys():
        if key not in names:
            raise KeyError(f'Invalid key - {repr(key)}.')


class BaseGenerator:
    """
    Base generator class.
    """

    def __init__(self, values, names: Optional[List[str]] = None):
        """
        Constructor of the :class:`hbutils.testing.BaseGenerator` class.

        :param values: Selection values, such as ``{'a': [2, 3], 'b': ['b', 'c']}``.
        :param names: Names of the given generator, default is ``None`` which means use the sorted \
            key set of the values.
        """
        self.__values = _single_dict_process(values)
        self.__names = list(names or sorted(self.__values.keys()))

    @property
    def values(self) -> Mapping[str, Tuple[object, ...]]:
        """
        Selection values.
        """
        return self.__values

    @property
    def names(self) -> List[str]:
        """
        Name of the given generator.
        """
        return self.__names

    def cases(self) -> Iterator[Mapping[str, object]]:
        """
        Virtual method for cases, will be implemented in child classes.
        """
        raise NotImplementedError  # pragma: no cover

    def tuple_cases(self) -> Iterator[Tuple[object, ...]]:
        """
        Tuple-formatted cases, can be used to iterate.
        """
        for d in self.cases():
            yield tuple(d[name] for name in self.names)



import socket
import warnings
from typing import Optional, Iterable

__all__ = [
    'is_free_port',
    'get_free_port',
]


def is_free_port(port: int) -> bool:
    """
    Overview:
        Check if given ``port`` is currently not in use and able to be allocated.

    :param port: Port to be checked.
    :return: In use or not.

    Examples::
        >>> from hbutils.system import is_free_port
        >>> is_free_port(22)
        False
        >>> is_free_port(80)
        False
        >>> is_free_port(8080)
        True
        >>> is_free_port(35022)
        True
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) != 0


def get_free_port(ports: Optional[Iterable[int]] = None, strict: bool = True) -> int:
    """
    Overview:
        Get allocatable port inside the given ``port``.

    :param ports: Range of ports to select.
    :param strict: Strictly use the ports in ``ports``. Default is ``False``, otherwise the ports \
        not in ``ports`` will probably be used.
    :param: A usable port.
    :raise OSError: Raise ``OSError`` when no ports can be allocated.

    Examples::
        >>> from hbutils.system import get_free_port
        >>> get_free_port(range(22, 1060))
        1024
        >>> get_free_port(range(1080, 2080, 10))  # assume that 1080 is in use
        1090
        >>> get_free_port(range(22, 80))
        OSError: No free port can be allocated with in range(22, 80).
        >>> get_free_port(range(22, 80), strict=False)
        44317

    .. note::
        Due to the safety consideration of OS, only ports over 1024 will be used.
    """
    _ports = [p for p in (ports or []) if p >= 1024]
    if not strict or not _ports:
        if not _ports and strict:
            warnings.warn('No usable ports provided, so strict mode will be disabled.', stacklevel=2)
            strict = False

    if _ports:
        for port in _ports:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                if s.connect_ex(('localhost', port)) != 0:
                    return port

    if not strict:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            s.bind(('', 0))
            _, port_ = s.getsockname()
            return port_
    else:
        raise OSError(f'No free port can be allocated with in {ports}.')



"""
Overview:
    Useful utilities for time units, such as h/m/s.
"""
import datetime
from typing import Union

from pytimeparse import parse as parse_duration

__all__ = ['time_to_duration', 'time_to_delta_str']

_TIME_TYPING = Union[float, int, str]


def time_to_duration(time_: _TIME_TYPING) -> Union[float, int]:
    """
    Overview:
        Turn any types of time duration in time value in seconds.

    Arguments:
        - time\\_ (:obj:`Union[int, float, str]`): Any types of time duration.

    Returns:
        - bytes (:obj:`int`): Time duration value in seconds.

    Examples::
        >>> from hbutils.scale import time_to_duration
        >>> time_to_duration(23344)
        23344
        >>> time_to_duration(233.54)
        233.54
        >>> time_to_duration('1h343m67.4s')
        24247.4
        >>> time_to_duration('0:03:53.540000')
        233.54
    """
    if isinstance(time_, (float, int)):
        return time_
    elif isinstance(time_, str):
        return parse_duration(time_)
    else:
        raise TypeError('{float}, {int} or {str} expected but {actual} found.'.format(
            float=float.__name__,
            int=int.__name__,
            str=str.__name__,
            actual=type(time_).__name__,
        ))


def time_to_delta_str(time_: _TIME_TYPING) -> str:
    """
    Overview:
        Turn any types of time duration in time value in formatted string.

    Arguments:
        - time\\_ (:obj:`Union[int, float, str]`): Any types of time duration.

    Returns:
        - bytes (:obj:`int`): Time duration value in formatted string.

    Examples::
        >>> from hbutils.scale import time_to_delta_str
        >>> time_to_delta_str(23344)
        '6:29:04'
        >>> time_to_delta_str(233.54)
        '0:03:53.540000'
        >>> time_to_delta_str('1h343m67.4s')
        '6:44:07.400000'
    """
    return str(datetime.timedelta(seconds=time_to_duration(time_)))



"""
Overview:
    Dimension operations.
"""
__all__ = [
    'cube_shape', 'dimension_switch', 'swap_2d',
]


def _s_cube_shape(c, path):
    if isinstance(c, (list, tuple)):
        n = len(c)
        if n == 0:
            return (0,)
        else:
            first = c[0]
            first_shape = _s_cube_shape(first, (*path, 0))
            for i, item in enumerate(c[1:], start=1):
                item_shape = _s_cube_shape(item, (*path, i))
                if first_shape != item_shape:
                    raise ValueError(f'Mismatching between {repr((*path, 0))} and {repr((*path, 1))}, '
                                     f'this is not a cube!', ((*path, 0), first_shape), ((*path, 1), item_shape))
            return (n, *first_shape)
    else:
        return ()


def cube_shape(c):
    """
    Overview:
        Get the shape of cube array.
        When it is not a cube, raise ``ValueError``.
    Arguments:
        - c: The given cube.

    Returns:
        - shape: Shape of the cube.

    Examples::
        >>> import numpy
        >>> from hbutils.collection import cube_shape
        >>> a = numpy.random.randint(-5, 15, (3, 5, 7, 9)).tolist()
        >>> cube_shape(a)
        (3, 5, 7, 9)
    """
    return _s_cube_shape(c, ())


def dimension_switch(c, dimensions):
    """
    Overview:
        Switch the dimensions of the cube array.
    Arguments:
        - c: Multiple dimensioned array.
        - dimensions: New order of dimensions, should be a tuple of 0 - N-1.

    Returns:
        - switched: Switched array.

    Examples::
        >>> import numpy
        >>> from hbutils.collection import cube_shape, dimension_switch
        >>> a = numpy.random.randint(-5, 15, (3, 5, 7, 9)).tolist()
        >>> cube_shape(a)
        (3, 5, 7, 9)
        >>> b = dimension_switch(a, (3, 0, 2, 1))
        >>> cube_shape(b)
        (9, 3, 7, 5)
    """
    shape = cube_shape(c)
    n = len(shape)
    if sorted(set(dimensions)) != list(range(len(shape))):
        raise ValueError(f'Invalid dimensions - {repr(dimensions)}.', dimensions)

    def _recursive(p, path):
        if p >= n:
            actual_path = [None] * n
            for i, di in enumerate(dimensions):
                actual_path[di] = path[i]

            oc = c
            for ip in actual_path:
                oc = oc[ip]
            return oc
        else:
            return [_recursive(p + 1, (*path, i)) for i in range(shape[dimensions[p]])]

    return _recursive(0, ())


def swap_2d(c):
    """
    Overview:
        Swap 2d array's dimension.
    Arguments:
        - c: 2d array.

    Returns:
        - swapped: Swapped array.

    Examples::
        >>> from hbutils.collection import swap_2d
        >>> swap_2d([
        ...     [9, 6, 4, 11, 5, -2, 1],
        ...     [0, 0, 11, 5, 8, -4, 9],
        ...     [0, 2, 13, 7, 0, 13, 0]
        ... ])
        [[9, 0, 0],
         [6, 0, 2],
         [4, 11, 13],
         [11, 5, 7],
         [5, 8, 0],
         [-2, -4, 13],
         [1, 9, 0]]
    """
    return dimension_switch(c, (1, 0))



"""
Overview:
    Useful utilities for memory size units, such as MB/KB/B.
"""
import warnings
from enum import IntEnum, unique
from typing import Union, Optional

from bitmath import Byte, NIST, SI
from bitmath import parse_string_unsafe as parse_bytes

from ..model import int_enum_loads

__all__ = [
    'size_to_bytes', 'size_to_bytes_str'
]

_EPS = 1e-10


def _is_int(value: Union[int, float], stacklevel: int = 4) -> int:
    if isinstance(value, float):
        _actual = int(round(value))
        _delta = abs(value - _actual)
        if _delta >= _EPS:
            warnings.warn(UserWarning(f'Float detected in variable in bytes ({repr(value)}), '
                                      f'rounded integer value ({repr(_actual)}) is used.'), stacklevel=stacklevel)
        return _actual
    elif isinstance(value, int):
        return value
    else:
        assert False, f"Should not reach here, {repr(type(value))} detected, " \
                      f"something may be wrong with {__name__}._is_int function."  # pragma: no cover


def _base_size_to_bytes(size, stacklevel: int = 4) -> int:
    if isinstance(size, (float, int)):
        return _is_int(size, stacklevel)
    elif isinstance(size, str):
        return _is_int(parse_bytes(size).bytes, stacklevel)
    elif isinstance(size, Byte):
        return _is_int(size.bytes, stacklevel)
    else:
        raise TypeError('{int}, {str} or {byte} expected but {actual} found.'.format(
            int=int.__name__,
            str=str.__name__,
            byte=Byte.__name__,
            actual=type(size).__name__,
        ))


_SIZE_TYPING = Union[int, float, str, Byte]


def size_to_bytes(size: _SIZE_TYPING) -> int:
    """
    Overview:
        Turn any types of memory size to integer value in bytes.

    Arguments:
        - size (:obj:`Union[int, float, str, Byte]`): Any types of size information.

    Returns:
        - bytes (:obj:`int`): Int formatted size in bytes.

    Examples::
        >>> from hbutils.scale import size_to_bytes
        >>> size_to_bytes(23344)
        23344
        >>> size_to_bytes('23356 KB')
        23356000
        >>> size_to_bytes('3.54 GB')
        3540000000
        >>> size_to_bytes('3.54 GiB')
        __main__:1: UserWarning: Float detected in variable in bytes (3801046056.96), rounded integer value (3801046057) is used.
        3801046057
    """
    return _base_size_to_bytes(size)


@int_enum_loads(name_preprocess=str.upper)
@unique
class SizeSystem(IntEnum):
    NIST = NIST
    SI = SI


def size_to_bytes_str(size: _SIZE_TYPING, precision: Optional[int] = None, system='nist') -> str:
    """
    Overview:
        Turn any types of memory size to string value in the best unit.

    :param size: Any types of size information.
    :param precision: Precsion for float values. Default is ``None`` which means just show the original float number.

    Returns:
        - bytes (:obj:`int`): String formatted size value in the best unit.

    Examples::
        >>> from hbutils.scale import size_to_bytes_str
        >>> size_to_bytes_str(23344)
        '22.796875 KiB'
        >>> size_to_bytes_str('23356 KB')
        '22.274017333984375 MiB'
        >>> size_to_bytes_str('3.54 GB')
        '3.296881914138794 GiB'
        >>> size_to_bytes_str('3.54 GiB')
        __main__:1: UserWarning: Float detected in variable in bytes (3801046056.96), rounded integer value (3801046057) is used.
        '3.540000000037253 GiB'
        >>> size_to_bytes_str('3.54 GB', precision=0)  # use precision
        '3 GiB'
        >>> size_to_bytes_str('3.54 GB', precision=3)
        '3.297 GiB'
        >>> size_to_bytes_str('3.54 GB', system='si')  # use GB/MB/KB instead of GiB/MiB/KiB
        '3.54 GB'
        >>> size_to_bytes_str('3.54 GB', precision=3, system='si')
        '3.540 GB'
    """
    system = SizeSystem.loads(system)
    if precision is None:
        format_str = "{value} {unit}"
    else:
        format_str = f"{{value:.{precision}f}} {{unit}}"
    return Byte(_base_size_to_bytes(size)).best_prefix(system.value).format(format_str)



from .base import Expression

__all__ = [
    'CheckExpression',
    'ComparableExpression',
    'IndexedExpression',
    'ObjectExpression',
    'LogicalExpression',
    'MathExpression',
    'BitwiseExpression',
]


class CheckExpression(Expression):
    """
    Overview:
        Check expression.

    Features:
        * ``__eq__``, which means ``x == y``.
        * ``__ne__``, which means ``x == y``.
    """

    def __eq__(self, other):
        return self._func(lambda x, y: x == y, self, other)

    def __ne__(self, other):
        return self._func(lambda x, y: x != y, self, other)


class ComparableExpression(CheckExpression):
    """
    Overview:
        Comparable expression.

    Features:
        * ``__eq__``, which means ``x == y`` (the same as :class:`CheckExpression`).
        * ``__ne__``, which means ``x == y`` (the same as :class:`CheckExpression`).
        * ``__ge__``, which means ``x >= y``.
        * ``__gt__``, which means ``x > y``.
        * ``__le__``, which means ``x <= y``.
        * ``__lt__``, which means ``x < y``.
    """

    def __le__(self, other):
        return self._func(lambda x, y: x <= y, self, other)

    def __lt__(self, other):
        return self._func(lambda x, y: x < y, self, other)

    def __ge__(self, other):
        return self._func(lambda x, y: x >= y, self, other)

    def __gt__(self, other):
        return self._func(lambda x, y: x > y, self, other)


class IndexedExpression(Expression):
    """
    Overview:
        Indexed expression.

    Features:
        * ``__getitem__``, which means ``x[y]``.
    """

    def __getitem__(self, item):
        return self._func(lambda x, y: x[y], self, item)


class ObjectExpression(Expression):
    """
    Overview:
        Object-like expression.

    Features:
        * ``__getattr__``, which means ``x.y``.
        * ``__call__``, which means ``x(*args, **kwargs)``.
    """

    def __getattr__(self, item):
        return self._func(lambda x, y: getattr(x, y), self, item)

    def __call__(self, *args, **kwargs):
        return self._func(lambda s, *args_, **kwargs_: s(*args_, **kwargs_), self, *args, **kwargs)


class LogicalExpression(Expression):
    """
    Overview:
        Logic expression.

    Features:
        * ``__and__``, which means ``x and y`` (written as ``x & y``).
        * ``__or__``, which means ``x or y`` (written as ``x | y``).
        * ``__invert__``, which means ``not x`` (written as ``~x``).

    .. note::
        Do not use this with :class:`BitwiseExpression`, or unexpected conflict will be caused.
    """

    def __and__(self, other):
        return self._func(lambda x, y: x and y, self, other)

    def __rand__(self, other):
        return self._func(lambda x, y: x and y, other, self)

    def __or__(self, other):
        return self._func(lambda x, y: x or y, self, other)

    def __ror__(self, other):
        return self._func(lambda x, y: x or y, other, self)

    def __invert__(self):
        return self._func(lambda x: not x, self)


class MathExpression(Expression):
    """
    Overview:
        Math calculation expression.

    Features:
        * ``__add__``, which means ``x + y``.
        * ``__sub__``, which means ``x - y``.
        * ``__mul__``, which means ``x * y``.
        * ``__truediv__``, which means ``x / y``.
        * ``__floordiv__``, which means ``x // y``.
        * ``__mod__``, which means ``x % y``.
        * ``__pow__``, which means ``x ** y``.
        * ``__pos__``, which means ``+x``.
        * ``__neg__``, which means ``-x``.
    """

    def __add__(self, other):
        return self._func(lambda x, y: x + y, self, other)

    def __radd__(self, other):
        return self._func(lambda x, y: x + y, other, self)

    def __sub__(self, other):
        return self._func(lambda x, y: x - y, self, other)

    def __rsub__(self, other):
        return self._func(lambda x, y: x - y, other, self)

    def __mul__(self, other):
        return self._func(lambda x, y: x * y, self, other)

    def __rmul__(self, other):
        return self._func(lambda x, y: x * y, other, self)

    def __truediv__(self, other):
        return self._func(lambda x, y: x / y, self, other)

    def __rtruediv__(self, other):
        return self._func(lambda x, y: x / y, other, self)

    def __floordiv__(self, other):
        return self._func(lambda x, y: x // y, self, other)

    def __rfloordiv__(self, other):
        return self._func(lambda x, y: x // y, other, self)

    def __mod__(self, other):
        return self._func(lambda x, y: x % y, self, other)

    def __rmod__(self, other):
        return self._func(lambda x, y: x % y, other, self)

    def __pow__(self, power):
        return self._func(lambda x, y: x ** y, self, power)

    def __rpow__(self, other):
        return self._func(lambda x, y: x ** y, other, self)

    def __pos__(self):
        return self._func(lambda x: +x, self)

    def __neg__(self):
        return self._func(lambda x: -x, self)


class BitwiseExpression(Expression):
    """
    Overview:
        Bitwise expression class.

    Features:
        * ``__and__``, which means ``x & y`` (bitwise).
        * ``__or__``, which means ``x | y`` (bitwise).
        * ``__xor__``, which means ``x ^ y`` (bitwise).
        * ``__lshift__``, which means ``x << y`` (bitwise).
        * ``__rshift__``, which means ``x >> y`` (bitwise).
        * ``__invert__``, which means ``~x`` (bitwise).

    .. note::
        Do not use this with :class:`LogicalExpression`, or unexpected conflict will be caused.
    """

    def __or__(self, other):
        return self._func(lambda x, y: x | y, self, other)

    def __ror__(self, other):
        return self._func(lambda x, y: x | y, other, self)

    def __xor__(self, other):
        return self._func(lambda x, y: x ^ y, self, other)

    def __rxor__(self, other):
        return self._func(lambda x, y: x ^ y, other, self)

    def __and__(self, other):
        return self._func(lambda x, y: x & y, self, other)

    def __rand__(self, other):
        return self._func(lambda x, y: x & y, other, self)

    def __invert__(self):
        return self._func(lambda x: ~x, self)

    def __lshift__(self, other):
        return self._func(lambda x, y: x << y, self, other)

    def __rlshift__(self, other):
        return self._func(lambda x, y: x << y, other, self)

    def __rshift__(self, other):
        return self._func(lambda x, y: x >> y, self, other)

    def __rrshift__(self, other):
        return self._func(lambda x, y: x >> y, other, self)



"""
Overview:
    Some useful utilities for decorator pattern and python decorators \
    for function or class.
"""
__all__ = ['decolize']

from functools import wraps, partial
from typing import Callable, Optional

from .singleton import SingletonMark

_NO_FUNC = SingletonMark('NO_FUNC_FOR_DECORATOR')


def decolize(deco: Callable[..., Callable]):
    """
    Overview:
        Decorator for decorator, make a decorator function with keyword-arguments \
        usable as the real python decorator.

    Arguments:
        - deco (:obj:`Callable[..., Callable]`): Decorator function to be decorated.

    Examples::

        >>> from functools import wraps
        >>> @decolize  # decorate the decorator
        ... def deco(func, a=1, b=2):
        ...     @wraps(func)
        ...     def _new_func(*args, **kwargs):
        ...         return func(*args, **kwargs) * a + b
        ...
        ...     return _new_func

        >>> @deco  # used as simple decorator, same as deco(_func_1)
        >>> def _func_1(a, b, c):
        ...     return (a + b * 2) * c

        >>> @deco(a=2)  # used as parameterized decorator, same as deco(_func_2, a=2)
        >>> def _func_2(a, b, c):
        ...     return (a + b * 2) * c

        >>> _func_1(1, 2, 3)
        17  # ((1 + 2 * 2) * 3) * 1 + 2 == 17
        >>> _func_2(1, 2, 3)
        32  # ((1 + 2 * 2) * 3) * 2 + 2 == 32

    .. tip::

        This decorator can be useful when build a decorator with parameters.

    .. note::

        In the decorated decorator, **only keyword arguments are supported**. \
        So make sure the original decorator function's parameter (except ``func``) are \
        all keyword supported, and do not try to use positional arguments when use it to \
        decorator another function.

    """

    @wraps(deco)
    def _new_deco(func: Optional[Callable] = _NO_FUNC, **kwargs):
        if func is _NO_FUNC:  # used as parameterized decorator
            return partial(deco, **kwargs)
        else:  # functional use
            return deco(func, **kwargs)

    return _new_deco



from enum import IntEnum, auto
from typing import List, Mapping, Union, Tuple, Optional

from .aetg import AETGGenerator
from .matrix import MatrixGenerator
from ...model import int_enum_loads
from ...reflection import progressive_for

__all__ = ['tmatrix']


@int_enum_loads(enable_int=False, name_preprocess=str.upper)
class MatrixMode(IntEnum):
    AETG = auto()
    MATRIX = auto()


def tmatrix(ranges: Mapping[Union[str, Tuple[str, ...]], List],
            mode='aetg', seed: Optional[int] = 0, level: int = 2) -> Tuple[List[str], List[Tuple]]:
    """
    Overview:
        Test matrix generator, which can be used in ``pytest.mark.parameterize``.

    :param ranges: Ranges of the values
    :param mode: Mode of the matrix, should be one of the ``aetg`` or ``matrix``. Default is ``aetg``.
    :param seed: Random seed, default is ``0`` which means the result is fixed (recommended).
    :param level: Lavel of AETG generating algorithm, default is ``2``.
    :returns: A tuple - ``(names, values)``.

    Examples::
        >>> from hbutils.testing import tmatrix
        >>> names, values = tmatrix(
        ...     {
        ...         'a': [2, 3],
        ...         'e': ['a', 'b', 'c'],
        ...         ('b', 'c'): [(1, 7), (4, 6), (9, 12)],
        ...     }
        ... )
        >>> print(names)
        ['a', 'e', 'b', 'c']
        >>> for i, v in enumerate(values):
        ...     print(i, v)
        0 (2, 'c', 9, 12)
        1 (3, 'c', 4, 6)
        2 (2, 'c', 1, 7)
        3 (3, 'b', 9, 12)
        4 (2, 'b', 4, 6)
        5 (3, 'b', 1, 7)
        6 (3, 'a', 9, 12)
        7 (2, 'a', 4, 6)
        8 (3, 'a', 1, 7)

    .. note::
        This can be directly used in ``pytest.mark.parametrize`` function.

        >>> @pytest.mark.unittest
        ... class TestTestingGeneratorFunc:
        ...     @pytest.mark.parametrize(*tmatrix({
        ...         'a': [2, 3],
        ...         'e': ['a', 'b', 'c'],
        ...         ('b', 'c'): [(1, 7), (4, 6), (9, 12)],
        ...     }))
        ...     def test_tmatrix_usage(self, a, e, b, c):
        ...         print(a, e, b, c)
    """
    mode = MatrixMode.loads(mode)

    key_map = {}
    final_names = []
    final_values = {}
    for ki, (key, value) in enumerate(ranges.items()):
        kname = f'key-{ki}'
        key_map[kname] = key
        final_names.append(kname)
        final_values[kname] = value

    names = []
    for key in ranges.keys():
        if isinstance(key, str):
            names.append(key)
        elif isinstance(key, tuple):
            for k in key:
                names.append(k)

    if mode == MatrixMode.MATRIX:
        generator = MatrixGenerator(final_values, final_names)
    elif mode == MatrixMode.AETG:
        generator = AETGGenerator(
            final_values, final_names, rnd=seed,
            pairs=list(progressive_for(final_names, min(level, len(names)))),
        )
    else:
        raise ValueError(f'Invalid mode - {mode!r}.')  # pragma: no cover

    pairs = []
    for case in generator.cases():
        _v_case = {}
        for name in final_names:
            key = key_map[name]
            if isinstance(key, str):
                _v_case[key] = case[name]
            elif isinstance(key, tuple):
                for ikey, ivalue in zip(key, case[name]):
                    _v_case[ikey] = ivalue

        pairs.append(tuple(_v_case[name] for name in names))

    return names, pairs



import platform

from packaging.version import Version

__all__ = [
    'python_version',
]


def python_version() -> Version:
    """
    Overview:
        Get version of python.

    :return: Version of python.

    Examples::
        >>> from hbutils.system import python_version
        >>>
        >>> python_version()
        Version('3.8.1')  # for example
    """
    return Version(platform.python_version())



"""
Overview:
    Encoding module, include some not so complex but useful functions. They will improve \
    the quality of your coding experience when dealing with these data.
"""
from .ansi import *
from .base64 import *
from .decode import *
from .hash import *



"""
Overview:
    Structural operations.
"""
from typing import Iterator, Tuple, Union, List

__all__ = [
    'sq_flatten',
    'nested_walk', 'nested_flatten',
]


def _g_sq_flatten(s):
    if isinstance(s, (list, tuple)):
        for item in s:
            yield from _g_sq_flatten(item)
    else:
        yield s


def sq_flatten(s):
    """
    Overview:
        Sequence flatten.
    Arguments:
        - s: The given sequence.

    Returns:
        - flatted: Flatted sequence.

    Examples::
        >>> from hbutils.collection import sq_flatten
        >>> sq_flatten([1, 2, [3, 4], [5, [6, 7], (8, 9, 10)], 11])
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
    """
    return list(_g_sq_flatten(s))


def _nested_walk(s, path):
    if isinstance(s, dict):
        for k, v in s.items():
            yield from _nested_walk(v, (*path, k))
    elif isinstance(s, (list, tuple)):
        for i, v in enumerate(s):
            yield from _nested_walk(v, (*path, i))
    else:
        yield path, s


def nested_walk(s) -> Iterator[Tuple[Tuple[Union[int, str], ...], object]]:
    """
    Overview:
        Walk for nested structure.

    Arguments:
        - s: Given structure.

    Returns:
        - walk_iter (:obj:`Iterator[Tuple[Tuple[Union[int, str], ...], object]]`): Iterator of the walk result.

    Examples::
        >>> from hbutils.collection import nested_walk
        >>> for p, v in nested_walk({'a': 1, 'b': ['c', 'd', {'x': (3, 4), 'y': 'f'}]}):
        ...     print(p, v)
        ...
        ('a',) 1
        ('b', 0) c
        ('b', 1) d
        ('b', 2, 'x', 0) 3
        ('b', 2, 'x', 1) 4
        ('b', 2, 'y') f
    """
    yield from _nested_walk(s, ())


def nested_flatten(s) -> List[Tuple[Tuple[Union[int, str], ...], object]]:
    """
    Overview:
        Flatten for nested structure.

    Arguments:
        - s: Given structure.

    Returns:
        - flatted (:obj:`List[Tuple[Tuple[Union[int, str], ...], object]]`): List of the flatted result.

    Examples::
        >>> from hbutils.collection import nested_flatten
        >>> print(nested_flatten({'a': 1, 'b': ['c', 'd', {'x': (3, 4), 'y': 'f'}]}))
        [(('a',), 1), (('b', 0), 'c'), (('b', 1), 'd'), (('b', 2, 'x', 0), 3), (('b', 2, 'x', 1), 4), (('b', 2, 'y'), 'f')]
    """
    return list(nested_walk(s))
